"""Classes to store peaks options."""

from peaks.core.utils.misc import format_colored_dict


class FileIOOptions:
    """Class used to set default options for file loading user-defined file paths, file extensions, and location.

    This class allows setting file paths, extensions, and location through its properties. Multiple paths and extensions
    can be provided as lists. The reset methods are used to reset specific options.

    Examples
    --------
    Example usage is as follows::

        import peaks as pks

        # Define file paths
        pks.opts.FileIO.path = ['sample1/i05-1-12', 'sample2/i05-1-12']

        # Define file extensions
        pks.opts.FileIO.ext = ['nxs', 'zip']

        # Define location
        pks.opts.FileIO.loc = 'Diamond I05-nano'

        # Show file options
        pks.opts.FileIO

        # Reset all FileIO optsions
        pks.opts.FileIO.reset()

        # Reset all options
        pks.opts.reset()
    """

    def __init__(self):
        self._path = None
        self._ext = None
        self._loc = None
        self._lazy_size = 1000000000  # Default lazy load size

    @property
    def path(self):
        return self._path

    @path.setter
    def path(self, value):
        if value is None:
            self._path = None
        elif isinstance(value, str) or (
            isinstance(value, list) and all(isinstance(p, str) for p in value)
        ):
            self._path = value
        else:
            raise TypeError(
                "Path must be a string or a list of strings pointing to desired file paths."
            )

    @path.deleter
    def path(self):
        self._path = None

    @property
    def ext(self):
        return self._ext

    @ext.setter
    def ext(self, value):
        if value is None:
            self._ext = None
        elif isinstance(value, str) or (
            isinstance(value, list) and all(isinstance(e, str) for e in value)
        ):
            self._ext = value
        else:
            raise TypeError(
                "Extension must be a string or a list of strings pointing to desired file extensions."
            )

    @ext.deleter
    def ext(self):
        self._ext = None

    @property
    def loc(self):
        return self._loc

    @loc.setter
    def loc(self, value):
        if value is None:
            self._loc = None
        elif isinstance(value, str):
            from peaks.core.fileIO.loc_registry import LOC_REGISTRY

            if value in LOC_REGISTRY:
                self._loc = value
            else:
                raise ValueError(
                    f"Location '{value}' is not in the list of available locations: {set(LOC_REGISTRY.keys())}"
                )
        else:
            raise TypeError(
                f"Location must be a string specifying an available file loader from {set(LOC_REGISTRY.keys())}."
            )

    @loc.deleter
    def loc(self):
        self._loc = None

    @property
    def lazy_size(self):
        return self._lazy_size

    @lazy_size.setter
    def lazy_size(self, value):
        if isinstance(value, int):
            self._lazy_size = value
        else:
            raise TypeError(
                "Lazy size must be an integer representing the file size (in bytes) for lazy loading."
            )

    @lazy_size.deleter
    def lazy_size(self):
        self._lazy_size = 1000000000  # Reset to default

    def reset(self):
        """Reset the file path, extension, location, and lazy size."""
        self._path = None
        self._ext = None
        self._loc = None
        self._lazy_size = 1000000000

    def __repr__(self):
        """Return a string representation of the current file variables."""
        return format_colored_dict(self.dict())

    def set(self, **kwargs):
        """Set the file path, extension, location, and lazy size.

        Parameters
        ----------
        kwargs : dict
            A dictionary of keyword arguments to set the file path, extension, location, and lazy size.
        """
        if "path" in kwargs:
            self.path = kwargs.pop("path")
        if "ext" in kwargs:
            self.ext = kwargs.pop("ext")
        if "loc" in kwargs:
            self.loc = kwargs.pop("loc")
        if "lazy_size" in kwargs:
            self.lazy_size = kwargs.pop("lazy_size")

        if kwargs:
            raise ValueError(
                f"Invalid keyword argument(s): {set(kwargs.keys())}. "
                f"Expected options from {set(self.dict().keys())}"
            )

    def dict(self):
        """Return a dictionary representation of the current file variables."""
        raw_dict = vars(self)
        return {k.lstrip("_"): raw_dict[k] for k in raw_dict}


class Options:
    """
    Singleton class to hold all fixed option groups, such as FileIO.

    This class provides a centralized way to manage various options used in the `peaks` package.
    It provides methods to reset, display, and access these options, and can be used with a context manager.

    Attributes
    ----------
    FileIO : FileIOOptions
        An instance of the FileIOOptions class to manage file input/output settings.

    Methods
    -------
    reset()
        Reset all option groups to their default values.

    Examples
    --------
    Example usage is as follows::

        import peaks as pks

        # Set some FileIO options
        pks.opts.FileIO.path = ['sample1/i05-1-12', 'sample2/i05-1-12']  # Default paths
        pks.opts.FileIO.ext = ['nxs', 'zip']  # Default extensions
        pks.opts.FileIO.loc = 'Diamond_I05_nano-ARPES'  # loc to use
        pks.opts.FileIO.lazy_size = 500000000  # Set lazy size to 500 Mb

        # Display all the current options
        pks.opts

        # Clear the location
        del pks.opts.FileIO.loc

        # Reset all options
        pks.opts.reset()

    Can also be used as a context manager to temporarily set options::

            import peaks as pks

            with pks.opts as opts:
                opts.FileIO.path = 'sads'
                opts.FileIO.loc = None
                opts.FileIO.ext = ['nxs', 'zip']
                opts.FileIO.lazy_size = 500000000

                # Display all the current options
                print(pks.opts)

            # Options are reset to their original state
            pks.opts
    """

    _instance = None
    _fileio_old_opts = None

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super(Options, cls).__new__(cls)
            cls._instance.FileIO = FileIOOptions()  # Initialize FileIO options here
        return cls._instance

    def reset(self):
        """Reset all option groups."""
        self.FileIO.reset()

    def dict(self):
        """Return a dictionary representation of the current options."""
        return {k: v.dict() for k, v in vars(self).items() if not k.startswith("_")}

    def __enter__(self):
        """Enter the context manager, storing the current state."""
        self._fileio_old_opts = self.FileIO.dict().copy()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit the context manager, resetting the state to the stored state."""
        if self._fileio_old_opts is not None:
            self.FileIO.set(**self._fileio_old_opts)
        self._fileio_old_opts = None

    def __repr__(self):
        return format_colored_dict(self.dict())


# Create the global `opts` instance
opts = Options()



"""Functions used to process data."""

# from .data_select import (
#     DC,
#     MDC,
#     EDC,
#     DOS,
#     tot,
#     radial_cuts,
#     extract_cut,
#     mask_data,
#     disp_from_hv,
# )
# from .k_conversion import k_convert
# from .fermi_level_correction import set_EF_correction
# from .differentiate import deriv, d2E, d2k, dEdk, dkdE, curvature, min_gradient
# from .tools import (
#     norm,
#     bgs,
#     bin_data,
#     smooth,
#     rotate,
#     sym,
#     sym_nfold,
#     degrid,
#     sum_data,
#     merge_data,
#     drift_correction,
# )
# from .angles import NormAngles



import numba
import numpy as np
from scipy.constants import physical_constants
from scipy.special import expit

TINY = 1.0e-15
kb_eV, _, _ = physical_constants["Boltzmann constant in eV/K"]


@numba.njit(cache=True)
def _linear_dos_fermi(
    x,
    EF,
    T,
    dos_slope,
    dos_intercept,
    bg_slope,
    bg_intercept,
):
    """Base function for fitting typical poly-Au Fermi edge data. Includes: Fermi cutoff (EF, T), linear DOS (dos_),
    and linear background above E_F accounting for e.g. inhomogeneous detector efficiency (bg_).

    Parameters:
    -----------
    x : np.ndarray
        Energy values in eV.
    EF : float
        The Fermi level in eV.
    T : float
        The temperature in K
    dos_slope : float
        The slope of the linear DOS below the Fermi level
    dos_intercept : float
        The intercept of the linear DOS below the Fermi level
    bg_slope : float
         The slope of the linear background above the Fermi level
    bg_intercept : float
        The intercept of the linear background above the Fermi level

    Returns
    -------
    numpy.ndarray
        The calculated Fermi function + DOS and backgrounds vs. x
    """

    return (bg_intercept + bg_slope * x) + (
        dos_intercept - bg_intercept + (dos_slope - bg_slope) * x
    ) / (1 + np.exp((1.0 * x - EF) / max(TINY, T * kb_eV)))


def _fermi_function(x, EF, T):
    """
    Fermi function in eV units.

    Parameters
    ----------
    x : np.ndarray or list
        Current energy value in eV.
    EF : float
        The Fermi level in eV.
    T : float
        The temperature in K.

    Returns
    -------
    numpy.ndarray
        The calculated Fermi function vs. x
    """

    return expit((EF - x) / max(kb_eV * T, TINY))



"""Master function for interactive display panels."""

import xarray as xr


def disp(data, primary_dim=None, exclude_from_centering="eV"):
    """GUI data viewer for 2D, 3D, or 4D data. Click on the Help menu of the GUI to see keyboard shortcuts.
    Default display mode is based on the conventional `peaks` data structure for ARPES data, but can be modified by
    passing the relevant parameters in the function call.

    Parameters
    ------------
    data : list or xarray.DataArray
         Either a single 2D, 3D, or 4D :class:`xarray.DataArray` or a list of 2D :class:`xarray.DataArray` objects.

    primary_dim : str or tuple of str or list of str, optional
        The primary dimension for the viewer, should be a single dim for 2D and 3D plots or a tuple or list of two
        dims for a 4D plot. Primary dim will be shown:
        - on the y-axis for 2D plots
        - as the central panel for 3D plots
        - as the primary data explorer panel for 4D plots
        Default behaviour is based on the conventional peaks data structure for ARPES data.

    exclude_from_centering : str or tuple of str or list of str or None, optional
        The dimension to exclude from centering for 2D and 3D data. Default is 'eV'.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # To display a single dispersion
        disp = load('disp.ibw')  # Load the dispersion
        disp.disp()  # Display it in the UI

        # To display a set of 2D dispersions with theta_par axis on the vertical axis
        disps_to_plot = [load(f"disp{i}.ibw") for i in range(1, 5)]  # Load dispersions
        pks.disp(disps_to_plot, primary_dim='theta_par')  # Open dispersions in viewer

        # To display a Fermi surface map
        FS_map = load('FS_map.nxs')
        FS_map.disp()

    """

    # If a DataTree instance is passed, try to parse into a form `.disp` can handle
    err_str = (
        "DataTree does not meet the required format to call `.disp`. It should either be a single leaf or a tree "
        "of multiple leaves each containing a Dataset with a single data variable of `data`. If a multi-leaf "
        "tree, only the 2D DataArrays are parsed."
    )
    if isinstance(data, xr.DataTree):
        if len(data.children) == 0:
            try:
                data = data.data
            except AttributeError as e:
                raise ValueError(err_str) from e
        else:
            try:
                data = [
                    node.get_DataArray()
                    for node in data.subtree
                    if not node.is_empty and len(node.data.dims) == 2
                ]
            except AttributeError as e:
                raise ValueError(err_str) from e

    if isinstance(data, list):
        for array in data:
            if isinstance(array, xr.DataArray):
                if len(array.dims) != 2:
                    raise ValueError(
                        "Unexpected dimensionality: only 2D xr.DataArrays are supported "
                        "for passing to `.disp` in a list."
                    )
            else:
                raise ValueError(
                    "Unexpected data type: only 2D xr.DataArrays are supported "
                    "for passing to `.disp` in a list."
                )

        from peaks.core.GUI.disp_panels.disp_2d import _disp_2d

        _disp_2d(
            data, primary_dim=primary_dim, exclude_from_centering=exclude_from_centering
        )

    elif isinstance(data, xr.DataArray):
        if 1 in data.shape:  # Check if there is a dummy axis that can be squeezed
            data = data.squeeze()
        ndim = len(data.dims)
        if ndim == 2:
            from peaks.core.GUI.disp_panels.disp_2d import _disp_2d

            _disp_2d(
                [data],
                primary_dim=primary_dim,
                exclude_from_centering=exclude_from_centering,
            )
        elif ndim == 3:
            from peaks.core.GUI.disp_panels.disp_3d import _disp_3d

            _disp_3d(
                data.pint.dequantify(),
                primary_dim=primary_dim,
                exclude_from_centering=exclude_from_centering,
            )
        elif ndim == 4:
            from peaks.core.GUI.disp_panels.disp_4d import _disp_4d

            _disp_4d(data.pint.dequantify(), primary_dim=primary_dim)
        else:
            raise ValueError(
                "Number of dimensions not supported for interactive display. "
                "Only 2D, 3D, or 4D xr.DataArrays are supported."
            )
    else:
        raise ValueError(
            "Data type not currently supported for interactive display. Only single 2D, 3D, or 4D xr.DataArrays or"
            "a list of 2D xr.DataArrays are supported."
        )



import xarray as xr

from peaks.core.accessors.accessor_methods import (
    PeaksDataTreeIteratorAccessor,
    register_lazy_accessor,
)

# List of function names (and the module they belong to) to directly register the accessor on DataTree
functions_to_register_for_direct_accessor = {
    "fileIO.data_saving": [
        "save",
    ],
    "GUI.disp_panels.disp": ["disp"],
    "utils.datatree_utils": ["view", "add_scan_group", "add", "get_DataArray"],
    "display.plotting": ["plot_grid", "plot_DCs"],
    "process.tools": ["sum_data", "subtract_data", "merge_data"],
}


# List of functions to register where the function should map over the tree
functions_to_register_for_iterable_accessor = {
    "process.data_select": [
        "DC",
        "MDC",
        "EDC",
        "DOS",
        "tot",
    ],
    "process.differentiate": [
        "deriv",
        "d2E",
        "d2k",
        "dEdk",
        "dkdE",
        "curvature",
        "min_gradient",
    ],
    "process.k_conversion": ["k_convert"],
    "process.tools": [
        "norm",
        "bgs",
        "bin_data",
        "bin_spectra",
        "smooth",
        "rotate",
        "sym",
        "sym_nfold",
        "degrid",
    ],
}


# Register each function as a lazy accessor on DataTree
for module_name, func_names in functions_to_register_for_direct_accessor.items():
    for func_name in func_names:
        accessor_property = register_lazy_accessor(
            func_name, f"peaks.core.{module_name}", xr.DataTree
        )

for module_name, func_names in functions_to_register_for_iterable_accessor.items():
    for func_name in func_names:
        accessor_property = register_lazy_accessor(
            func_name,
            f"peaks.core.{module_name}",
            xr.DataTree,
            accessor_class=PeaksDataTreeIteratorAccessor,
        )



def _parse_norm_emission_cursor_stats(da, norm_emission_dict):
    cursor_text = ""
    if norm_emission_dict:
        cursor_text += "Normal emission: "
        # Try and parse local names
        for dim in norm_emission_dict.copy():
            local_name = _get_nested_attr(
                da.metadata, f"manipulator.{dim}.local_name"
            ) or _get_nested_attr(da.metadata, f"analyser.deflector.{dim}.local_name")
            if local_name:
                norm_emission_dict[f"{dim} [{local_name}]"] = norm_emission_dict.pop(dim)

        # Set the normal emission text
        cursor_text += ", ".join([f"{k}={v:.2f}" for k, v in norm_emission_dict.items()])
    else:
        cursor_text = "<br>"
    cursor_text += "<hr>"

    return cursor_text


def _get_nested_attr(obj, attr_path):
    """Retrieve a nested attribute from an object using a dot-separated path."""
    attrs = attr_path.split(".")
    for attr in attrs:
        if hasattr(obj, attr):
            obj = getattr(obj, attr)
        else:
            return None
    return obj



"""Functions for the 3D interactive display panel."""

import sys
from functools import partial

import numpy as np
import pint
import pyperclip
import pyqtgraph as pg
from PyQt6 import QtCore, QtGui, QtWidgets
from PyQt6.QtWidgets import (
    QApplication,
    QHBoxLayout,
    QLabel,
    QVBoxLayout,
    QWidget,
)

from peaks.core.GUI.GUI_utils import (
    Crosshair,
    KeyPressGraphicsLayoutWidget,
)
from peaks.core.GUI.GUI_utils.cursor_stats import _parse_norm_emission_cursor_stats
from peaks.core.metadata.metadata_methods import display_metadata
from peaks.core.process.tools import estimate_sym_point, sym


def _disp_3d(data, primary_dim, exclude_from_centering):
    """Display a 3D interactive display panel.

    Parameters
    ------------
    data : xarray.DataArray
         A single 3D :class:`xarray.DataArray`.

    primary_dim : str
        The primary dimension for the viewer, used to select the plane shown in the central panel.

    exclude_from_centering : str or tuple of str or list of str or None
        The dimension to exclude from centering. Default is 'eV'.

    """
    global app  # Ensure the QApplication instance is not garbage collected
    app = QApplication.instance()
    if app is None:
        app = QApplication(sys.argv)

    viewer = _Disp3D(data, primary_dim, exclude_from_centering)
    viewer.show()
    app.exec()


class _Disp3D(QtWidgets.QMainWindow):
    def __init__(self, data, primary_dim, exclude_from_centering):
        super().__init__()

        # Parse data
        if primary_dim:  # UI functions assume primary dim is penultimate dim
            primary_dim_index = data.dims.index(primary_dim)
            if primary_dim_index != 1:
                dims = list(data.dims)
                dims.remove(primary_dim)
                dims.insert(1, primary_dim)
                data = data.transpose(*dims)
        self.data = data.compute()

        # Read scan metadata
        self.metadata_text = "<span style='color:white'>"
        self.metadata_text += display_metadata(self.data, "html")
        self.metadata_text += "</span><br>"

        # Crosshair options
        self.DC_pen = (238, 119, 51)
        self.DC_mirror_pen = pg.mkPen(
            color=(51, 187, 238),
            style=QtCore.Qt.PenStyle.DashLine,
        )
        self.align_aid_pen = pg.mkPen(
            color=(204, 51, 153),
            style=QtCore.Qt.PenStyle.DotLine,
            width=2.5,
        )
        self.xh_brush = (238, 119, 51, 60)
        self.DC_xh_brush = (238, 119, 51, 80)

        # Set keys for the keyboard control
        def _get_key(param):
            return getattr(QtCore.Qt.Key, f"Key_{param}")

        self.key_modifiers_characters = {
            "hide_all": "Space",
            "move_primary_dim": "Shift",
            "show_hide_mirror": "M",
        }
        self.key_modifiers = {
            k: _get_key(v) for k, v in self.key_modifiers_characters.items()
        }
        self.move_primary_dim_key_enabled = False

        # Set options for dims to exclude from centering
        if isinstance(exclude_from_centering, str):
            self.exclude_from_centering = [exclude_from_centering]
        elif isinstance(exclude_from_centering, (tuple, list)):
            self.exclude_from_centering = list(exclude_from_centering)
        else:
            self.exclude_from_centering = []

        self._init_data()  # Read some basic parameters from the data
        self._init_UI()  # Initialize the GUI layout
        self._set_data()  # Initialize with data

        # Ensure the application quits when the window is closed
        self.setAttribute(QtCore.Qt.WidgetAttribute.WA_DeleteOnClose)
        self.closeEvent = self._close_application

    def _close_application(self, event):
        """Close the application when the window is closed."""
        self.graphics_layout.close()
        app.quit()

    # ##############################
    # GUI layout
    # ##############################
    def _init_UI(self):
        self.setWindowTitle("Display Panel")
        self.setGeometry(100, 100, 1200, 700)

        # Main window layout
        central_widget = QWidget()
        self.setCentralWidget(central_widget)
        layout = QHBoxLayout()
        central_widget.setLayout(layout)

        self._build_plot_layout(layout)  # Build the basic plot layout
        self._build_controls_layout(layout)  # Build the control panel layout
        self._build_menu()  # Add the menu

    def _build_plot_layout(self, layout):
        # Create a GraphicsLayoutWidget
        self.graphics_layout = (
            KeyPressGraphicsLayoutWidget._KeyPressGraphicsLayoutWidget()
        )
        self.graphics_layout.viewport().setAttribute(
            QtCore.Qt.WidgetAttribute.WA_AcceptTouchEvents, False
        )
        layout.addWidget(self.graphics_layout)

        # Set up the plots with the following number conventions
        #    #   --> indexes in array
        #   (..) --> dims displayed
        #        +----------+
        #        |  DC1 (2) |
        #        +----------+
        # +-----++----------++----------+
        # | DC0 ||  IMAGE 1 ||  IMAGE 2 |
        # | (0) ||   (0,2)  ||    (0,1) |
        # +-----++----------++----------+
        #        +----------+
        #        |  IMAGE 0 |
        #        |   (1,2)  |
        #        +----------+

        # Main image plots
        self.image_plots = []
        self.image_plots.append(self.graphics_layout.addPlot(2, 1, 1, 1))
        self.image_plots.append(self.graphics_layout.addPlot(1, 1, 1, 1))
        self.image_plots.append(self.graphics_layout.addPlot(1, 2, 1, 1))

        self.DC_plots = []
        self.DC_plots.append(self.graphics_layout.addPlot(1, 0, 1, 1))
        self.DC_plots.append(self.graphics_layout.addPlot(0, 1, 1, 1))

        self.DC_plots[0].setYLink(self.image_plots[1])
        self.DC_plots[1].setXLink(self.image_plots[1])
        self.image_plots[0].setXLink(self.image_plots[1])
        self.image_plots[2].setYLink(self.image_plots[1])
        for i in [0, 1]:
            self.DC_plots[i].getAxis("bottom").hide()
            self.DC_plots[i].getAxis("top").show()
        self.image_plots[2].getAxis("left").hide()
        self.image_plots[2].getAxis("right").show()
        for i in [1, 2]:
            self.image_plots[i].getAxis("bottom").hide()
            self.image_plots[i].getAxis("top").show()

        self.graphics_layout.ci.layout.setRowStretchFactor(1, 4)
        self.graphics_layout.ci.layout.setRowStretchFactor(2, 4)
        self.graphics_layout.ci.layout.setColumnStretchFactor(1, 4)
        self.graphics_layout.ci.layout.setColumnStretchFactor(2, 4)

        # Set margins to line up the plots
        self.DC_plots[0].getAxis("left").setWidth(40)
        self.DC_plots[0].getAxis("bottom").setWidth(50)
        self.DC_plots[1].getAxis("left").setWidth(60)
        self.image_plots[0].getAxis("left").setWidth(60)
        self.image_plots[1].getAxis("left").setWidth(60)
        self.image_plots[2].getAxis("right").setWidth(50)

    def _build_controls_layout(self, layout):
        # Right panel -------------------------------------
        right_panel_layout = QVBoxLayout()
        layout.addLayout(right_panel_layout)

        # Cursor stats etc.
        scroll_area = QtWidgets.QScrollArea()
        scroll_area.setWidgetResizable(True)
        scroll_area.setFixedHeight(150)
        self.cursor_stats = QLabel()
        self.cursor_stats.setStyleSheet(
            "QLabel { background-color : black; color : white; }"
        )
        self.cursor_stats.setAlignment(
            QtCore.Qt.AlignmentFlag.AlignTop | QtCore.Qt.AlignmentFlag.AlignLeft
        )
        self.cursor_stats.setWordWrap(True)
        self.cursor_stats.setMaximumWidth(450)
        self.cursor_stats.setMinimumWidth(300)
        scroll_area.setWidget(self.cursor_stats)
        right_panel_layout.addWidget(scroll_area)

        # xh positions and widths
        xh_group = QtWidgets.QGroupBox("Data selection")
        xh_group.setContentsMargins(5, 22, 5, 5)
        xh_group.setMaximumWidth(450)
        xh_layout = QtWidgets.QVBoxLayout()
        xh_layout.setContentsMargins(5, 0, 5, 0)
        xh_group.setLayout(xh_layout)
        right_panel_layout.addWidget(xh_group)
        # Add dimension labels positions and widths
        self.cursor_positions_selection = []
        self.cursor_widths_selection = []
        for i in range(3):
            # Make an HBox for each dimension
            hbox = QHBoxLayout()
            xh_layout.addLayout(hbox)
            label = QLabel(f"{self.dims[i]}")
            label.setFixedWidth(75)
            hbox.addWidget(label)
            # Make a selection box for cursor positions
            self.cursor_positions_selection.append(QtWidgets.QDoubleSpinBox())
            self.cursor_positions_selection[i].setRange(
                self.ranges[i][0], self.ranges[i][1]
            )
            self.cursor_positions_selection[i].setSingleStep(self.step_sizes[i])
            self.cursor_positions_selection[i].setDecimals(3)
            self.cursor_positions_selection[i].setFixedWidth(75)
            hbox.addWidget(self.cursor_positions_selection[i])
            # Make a selection box for cursor widths
            hbox.addWidget(QLabel("Δ:"))
            self.cursor_widths_selection.append(QtWidgets.QDoubleSpinBox())
            self.cursor_widths_selection[i].setRange(0, self.data_span[i])
            self.cursor_widths_selection[i].setSingleStep(self.step_sizes[i])
            self.cursor_widths_selection[i].setDecimals(3)
            self.cursor_widths_selection[i].setFixedWidth(75)
            hbox.addWidget(self.cursor_widths_selection[i])
            hbox.addStretch()
        self.mirror_checkbox = QtWidgets.QCheckBox("Mirror DCs?")
        xh_layout.addWidget(self.mirror_checkbox)

        right_panel_layout.addSpacing(10)

        # Alignment options
        align_group = QtWidgets.QGroupBox("Alignment")
        align_group.setContentsMargins(5, 22, 5, 5)
        align_group.setMaximumWidth(450)
        align_layout = QtWidgets.QHBoxLayout()
        align_layout.setContentsMargins(5, 0, 5, 0)
        align_group.setLayout(align_layout)
        right_panel_layout.addWidget(align_group)
        # Add alignment buttons
        vbox = QVBoxLayout()
        align_layout.addLayout(vbox)
        hbox = QHBoxLayout()
        hbox.setSpacing(5)
        vbox.addLayout(hbox)
        label = QLabel("Rotation:")
        label.setFixedWidth(90)
        hbox.addWidget(label)
        self.rotation_selection = QtWidgets.QDoubleSpinBox()
        self.rotation_selection.setRange(-180, 180)
        self.rotation_selection.setSingleStep(1)
        self.rotation_selection.setDecimals(2)
        self.rotation_selection.setFixedWidth(75)
        hbox.addWidget(self.rotation_selection)
        self.rotation_delta = [-45, -30, 30, 45]
        self.rotation_delta_buttons = []
        for ang in self.rotation_delta:
            button_text = f"{'+' if ang > 0 else '-'}{abs(ang)}°"
            button = QtWidgets.QPushButton(button_text)
            button.setFixedWidth(40)
            self.rotation_delta_buttons.append(button)
            hbox.addWidget(button)
        hbox.addStretch()
        hbox = QHBoxLayout()
        vbox.addLayout(hbox)
        label = QLabel("Alignment aid:")
        label.setFixedWidth(85)
        hbox.addWidget(label)
        self.alignment_aid = QtWidgets.QComboBox()
        self.alignment_aid.addItems(
            ["None", "Square", "Hexagon", "Hexagon (r30)", "Circle"]
        )
        self.alignment_aid_scaling = (self.ranges[2][1] - self.ranges[2][0]) / 150
        hbox.addWidget(self.alignment_aid)
        # Add a slider to set size of alignment aid
        self.alignment_aid_size = QtWidgets.QSlider(QtCore.Qt.Orientation.Horizontal)
        self.alignment_aid_size.setRange(0, 100)
        self.alignment_aid_size.setValue(50)
        self.alignment_aid_size.setSingleStep(1)
        self.alignment_aid_size.setFixedWidth(100)
        hbox.addWidget(self.alignment_aid_size)
        hbox.addStretch()

        hbox = QHBoxLayout()
        vbox.addLayout(hbox)
        self.align_button = QtWidgets.QPushButton("Align")
        hbox.addWidget(self.align_button)
        self.copy_button = QtWidgets.QPushButton("Copy")
        hbox.addWidget(self.copy_button)

        right_panel_layout.addSpacing(10)

        # Add colorbar
        colorbar_group = QtWidgets.QGroupBox("Image contrast")
        colorbar_group.setMaximumWidth(450)
        right_panel_layout.addWidget(colorbar_group)
        colorbar_container = QVBoxLayout()
        colorbar_group.setLayout(colorbar_container)
        self.colorbar_widget_container = pg.GraphicsLayoutWidget()
        self.colorbar_widget_container.viewport().setAttribute(
            QtCore.Qt.WidgetAttribute.WA_AcceptTouchEvents, False
        )
        self.colorbar_widget_container.setMaximumHeight(100)
        self.colorbar_widget_container.setMaximumWidth(400)
        colorbar_container.addWidget(self.colorbar_widget_container)
        self.cscale_lock = QtWidgets.QCheckBox("cScale locked", checked=True)
        colorbar_container.addWidget(self.cscale_lock)

        right_panel_layout.addStretch()

    def _build_menu(self):
        """Add a menu"""
        self.menu = self.menuBar().addMenu("Display Panel")
        self.shortcuts_action = QtGui.QAction("Help", self)

        self.help_text = f"""
                <table style='width:100%; border: 1px solid black; border-collapse: collapse;'>
                  <tr style='border: 1px solid black;'>
                    <th style='border: 1px solid black; padding: 5px; font-weight: normal;'>Action</th>
                    <th style='border: 1px solid black; padding: 5px; font-weight: normal;'>Key Binding</th>
                  </tr>
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Move crosshair in main plot</td>
                    <td style='padding: 5px; font-weight: normal;'>Arrow keys</td>
                  </tr>
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Change primary slice</td>
                    <td style='padding: 5px; font-weight: normal;'>{self.key_modifiers_characters["move_primary_dim"]}
                    + Up/Down arrow keys</td>
                  </tr>
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Hide all crosshairs</td>
                    <td style='padding: 5px; font-weight: normal;'>{self.key_modifiers_characters["hide_all"]}</td>
                  </tr>
                """

        self.help_text += f"""
                  <tr>
                    <td style='padding: 5px; font-weight: normal;'>Enable/disable mirrored mode</td>
                    <td style='padding: 5px; font-weight: normal;'>{self.key_modifiers_characters["show_hide_mirror"]}</td>
                  </tr>
                </table>
                """

        self.shortcuts_action.triggered.connect(
            lambda: QtWidgets.QMessageBox.information(self, "Help", self.help_text)
        )
        self.menu.addAction(self.shortcuts_action)

    # ##############################
    # Data handling / plotting
    # ##############################
    def _init_data(self):
        """Extract some core parameters from the data."""
        self.dims = self.data.dims
        self.coords = [self.data.coords[dim].values for dim in self.dims]
        self.step_sizes = [
            (coord[-1] - coord[0]) / (len(coord) - 1) for coord in self.coords
        ]
        self.ranges = [
            (min(coords) - self.step_sizes[i] / 2, max(coords) + self.step_sizes[i] / 2)
            for i, coords in enumerate(self.coords)
        ]
        self.data_span = [abs(range[1] - range[0]) for range in self.ranges]
        self.c_min = float(self.data.min())
        self.c_max = float(self.data.max())

        # Set centering dims
        self.centering_dims = [
            dim for dim in self.dims if dim not in self.exclude_from_centering
        ]

        # Attempt to read some metadata
        self.metadata_text = "<span style='color:white'>"
        self.metadata_text += display_metadata(self.data, "html")
        self.metadata_text += "</span><br>"

    def _set_main_plots(self):
        # Make arrays for plotitems and crosshairs
        self.image_items = []
        self.xhs = []

        # Populate these for each of the main plots
        for i in range(3):
            self.image_items.append(
                pg.ImageItem(self.images[i].values, axisOrder="row-major")
            )
            self.image_plots[i].addItem(self.image_items[i])

            # Transform to match data coordinates
            active_dim_nos = self._get_active_dim_nos(i)
            self.image_items[i].setTransform(
                pg.QtGui.QTransform(
                    self.step_sizes[active_dim_nos[1]],
                    0,
                    0,
                    0,
                    self.step_sizes[active_dim_nos[0]],
                    0,
                    self.ranges[active_dim_nos[1]][0],
                    self.ranges[active_dim_nos[0]][0],
                    1,
                )
            )

            # Crosshairs
            pos = tuple(
                [self.cursor_positions_selection[i].value() for i in active_dim_nos]
            )
            self.xhs.append(
                Crosshair._Crosshair(
                    self.image_plots[i],
                    pos=pos,
                    dim0_width=self.cursor_widths_selection[active_dim_nos[0]].value(),
                    dim1_width=self.cursor_widths_selection[active_dim_nos[1]].value(),
                    brush=self.xh_brush,
                    bounds=[
                        self.ranges[active_dim_nos[0]],
                        self.ranges[active_dim_nos[1]],
                    ],
                    axisOrder="row-major",
                    target_item_kwargs={
                        "pen": pg.mkPen(color=(238, 119, 51)),
                        "brush": pg.mkBrush(color=(238, 119, 51, 50)),
                    },
                    linear_region_item_kwargs={"pen": pg.mkPen(color=(238, 119, 51))},
                )
            )

        # Add alignment xh
        self.align_xhs = []
        for angle in [0, 90]:
            xh = pg.InfiniteLine(
                pos=(
                    self.cursor_positions_selection[2].value(),
                    self.cursor_positions_selection[0].value(),
                ),
                pen=self.align_aid_pen,
                angle=angle,
            )
            self.align_xhs.append(xh)
            self.image_plots[1].addItem(xh)
            xh.setVisible(False)

        # Add alignment aid
        dummy_centre = QtCore.QPointF(0, 0)
        d = self._create_alignment_shape(
            "Square",
            dummy_centre,
            self.alignment_aid_size.value() / 20,
            self.rotation_selection.value(),
        )
        self.alignment_aid_shape = pg.QtWidgets.QGraphicsPathItem(d)
        self.alignment_aid_shape.setPen(self.align_aid_pen)
        self.alignment_aid_shape.setVisible(False)
        self.image_plots[1].addItem(self.alignment_aid_shape)

        # Add colour bar to main image
        self.cmap = pg.colormap.get("Greys", source="matplotlib")
        self.colorbar = pg.ColorBarItem(
            label=self.data.name,
            colorMap=self.cmap,
            colorMapMenu=True,
            limits=(self.c_min, self.c_max),
            rounding=min(abs(self.c_max - self.c_min) / 2000, 1),
            interactive=True,
            orientation="h",
            values=(self.c_min, self.c_max),
        )
        self.colorbar.setImageItem(self.image_items)
        self.colorbar_widget_container.addItem(self.colorbar, row=0, col=0)

        # Flip axes as required
        self.image_plots[2].getViewBox().invertX(True)

    def _set_DC_plots(self):
        self.DC_plot_items = {}
        self.DC_plots_xhs = []
        for i in range(2):
            dim_no = [0, 2][i]  # Select the active dim for the relevant xh plot
            self.DC_plot_items[i] = []
            self.DC_plot_items[f"{i}_m"] = []
            DC = self._select_DC(dim_no)

            # Add to plots - get ordering correct
            if i == 0:
                a, b = DC.data, DC.coords[self.dims[dim_no]].values
            else:
                a, b = DC.coords[self.dims[dim_no]].values, DC.data

            self.DC_plot_items[i].append(
                self.DC_plots[i].plot(
                    a,
                    b,
                    pen=self.DC_pen,
                )
            )

            # Add spans for crosshairs
            self.DC_plots_xhs.append(
                pg.LinearRegionItem(
                    values=getattr(self.xhs[1], f"get_dim{i}_span")(),
                    orientation="horizontal" if i == 0 else "vertical",
                    pen=(0, 0, 0, 0),
                    brush=self.DC_xh_brush,
                    movable=False,
                )
            )
            self.DC_plots[i].addItem(self.DC_plots_xhs[i])

        self.DC_plots[0].getViewBox().invertX(True)

    def _set_plot_range_limits(self):
        # Set limits for image plots
        for i in range(3):
            active_dim_nos = self._get_active_dim_nos(i)
            xmin, xmax = self.ranges[active_dim_nos[1]]
            ymin, ymax = self.ranges[active_dim_nos[0]]
            self.image_plots[i].getViewBox().setLimits(
                xMin=xmin, xMax=xmax, yMin=ymin, yMax=ymax
            )

        # Set limits for DC plots
        ymin, ymax = self.ranges[0]
        xmin, xmax = self.ranges[2]
        self.DC_plots[0].getViewBox().setLimits(yMin=ymin, yMax=ymax)
        self.DC_plots[1].getViewBox().setLimits(xMin=xmin, xMax=xmax)

    def _set_plot_labels(self):
        # Extract labels
        dim_labels = []
        for i in range(3):
            dim_units = self.data.coords[self.dims[i]].attrs.get("units")
            dim_labels.append(
                f"{self.dims[i]} ({dim_units})" if dim_units else self.dims[i]
            )

        # Label the relevant plots
        self.DC_plots[0].setLabel("left", dim_labels[0])
        self.DC_plots[0].setLabel("top", " ")
        self.DC_plots[1].setLabel("top", dim_labels[2])
        self.image_plots[0].setLabel("bottom", dim_labels[2])
        self.image_plots[0].setLabel("left", dim_labels[1])
        self.image_plots[1].setLabel("top", dim_labels[2])
        self.image_plots[1].setLabel("left", dim_labels[0])
        self.image_plots[2].setLabel("top", dim_labels[1])
        self.image_plots[2].setLabel("right", dim_labels[0])

    def _select_DC(self, dim_no):
        """Select a DC along dim_no 0 from the data, handling averaging."""
        sum_over_dim_nos = [i for i in range(3) if i != dim_no]
        DC = self.data
        for i in sum_over_dim_nos:
            _pos = self.cursor_positions_selection[i].value()
            _width = self.cursor_widths_selection[i].value() / 2
            _range = slice(_pos - _width, _pos + _width)
            if _width < self.step_sizes[i]:
                DC = DC.sel({self.dims[i]: _pos}, method="nearest")
            else:
                DC = DC.sel({self.dims[i]: _range}).mean(self.dims[i])
        return DC

    # ##############################
    # Data / plot updates
    # ##############################
    def _set_data(self):
        """Set the data in the plots"""

        self.images = [None, None, None]

        # Initialise crosshair positions and widths - default to centre of data and 1/200th of range
        for i in range(3):
            self.cursor_positions_selection[i].setValue(sum(self.ranges[i]) / 2)
            self.cursor_widths_selection[i].setValue(self.data_span[i] / 200)

        # If eV in data, attempt to set a Fermi level and use this as initial xh pos
        if "eV" in self.dims:
            eV_dim = self.data.dims.index("eV")
            if (
                self.data.attrs.get("scan_type") == "hv scan"
                and self.data.attrs.get("eV_type") == "kinetic"
            ):
                # Due to the particular data structure of an hv scan, take the central slice to estimate EF
                hv = self.data.hv.data
                middle_hv = hv[(len(hv) - 1) // 2]
                EF = self.data.disp_from_hv(middle_hv).estimate_EF()
            else:
                EF = self.data.estimate_EF()
            if isinstance(EF, (list, np.ndarray)):
                if any(EF):
                    EF = sum(self.ranges[eV_dim]) / 2
            else:
                EF = EF

            self.cursor_positions_selection[eV_dim].setValue(
                EF or sum(self.ranges[eV_dim]) / 2
            )

        # Make primary data slice
        self._set_slice(1)

        # Try to estimate data centre from this slice and update other xh positions
        try:
            centre = estimate_sym_point(self.images[1], dims=self.centering_dims)
            for dim, coord in centre.items():
                self.cursor_positions_selection[self.dims.index(dim)].setValue(coord)
        except Exception:
            pass
        # Make the coresponding slices
        self._set_slice(0)
        self._set_slice(2)

        # Make plots
        self._set_main_plots()  # Make main plots
        self._set_DC_plots()  # Make DC plots
        self._set_plot_range_limits()  # Set plot range limits
        self._set_plot_labels()  # Set labels
        self._update_cursor_stats_text()  # Update initial cursor stats

        # Connect signals
        self._connect_signals_crosshairs()
        self._connect_signals_cursor_boxes_change()
        self._connect_signals_xh_span_change()
        self._connect_signals_align()
        self._connect_signals_mirror_checkbox()
        self._connect_key_press_signals()

    def _set_slice(self, dim_no):
        """Set a data slice based on the xh positions and widths"""
        _pos = self.cursor_positions_selection[dim_no].value()
        _width = self.cursor_widths_selection[dim_no].value() / 2
        _range = slice(_pos - _width, _pos + _width)
        if _width < self.step_sizes[dim_no]:
            self.images[dim_no] = self.data.sel(
                {self.dims[dim_no]: _pos}, method="nearest"
            )
        else:
            self.images[dim_no] = self.data.sel({self.dims[dim_no]: _range}).mean(
                dim=self.dims[dim_no]
            )

    def _update_cursor_boxes(self, xh_no):
        """Update the cursor boxes when the crosshair is moved."""
        x, y = self.xhs[xh_no].get_pos()
        active_dim_nos = self._get_active_dim_nos(xh_no)
        self.cursor_positions_selection[active_dim_nos[0]].setValue(x)
        self.cursor_positions_selection[active_dim_nos[1]].setValue(y)

    def _update_xh_positions(self, dim_changed):
        """Update the crosshair positions when the cursor boxes are changed."""
        # Update the crosshair positions on the main plots
        for i in range(3):
            active_dim_nos = self._get_active_dim_nos(i)
            if dim_changed in active_dim_nos:
                xh = self.xhs[i]
                xh.set_pos(
                    (
                        self.cursor_positions_selection[active_dim_nos[0]].value(),
                        self.cursor_positions_selection[active_dim_nos[1]].value(),
                    )
                )

        if dim_changed in [0, 2]:
            # Update the alignment crosshairs
            for i, xh in enumerate(self.align_xhs):
                xh.setPos(
                    (
                        self.cursor_positions_selection[2].value(),
                        self.cursor_positions_selection[0].value(),
                    )
                )
                xh.setAngle(self.rotation_selection.value() + i * 90)
            if self.alignment_aid_shape.isVisible():
                self._update_align_shape()

        # Update the crosshair positions on the DCs
        if dim_changed == 0:
            xh = self.DC_plots_xhs[0]
            xh.setRegion(self.xhs[1].get_dim0_span())
        elif dim_changed == 2:
            xh = self.DC_plots_xhs[1]
            xh.setRegion(self.xhs[1].get_dim1_span())

    def _update_xh_widths(self, dim_changed):
        """Update the crosshair widths when the cursor boxes are changed."""
        # Update the crosshairs on the main plots
        for i in range(3):
            active_dim_nos = self._get_active_dim_nos(i)
            if dim_changed in active_dim_nos:
                xh = self.xhs[i]
                xh.set_dim0_width(
                    self.cursor_widths_selection[active_dim_nos[0]].value()
                )
                xh.set_dim1_width(
                    self.cursor_widths_selection[active_dim_nos[1]].value()
                )

        # Update the crosshairs on the DCs
        if dim_changed == 0:
            xh = self.DC_plots_xhs[0]
            xh.setRegion(self.xhs[1].get_dim0_span())
        elif dim_changed == 2:
            xh = self.DC_plots_xhs[1]
            xh.setRegion(self.xhs[1].get_dim1_span())

    def _update_align_xh_angle(self):
        """Update the alignment crosshair angles when the rotation is changed."""
        for i, xh in enumerate(self.align_xhs):
            xh.setAngle(self.rotation_selection.value() + i * 90)
            if self.rotation_selection.value() == 0:
                xh.setVisible(False)
            else:
                xh.setVisible(True)

        if self.alignment_aid_shape.isVisible():
            self._update_align_shape()

        self._update_cursor_stats_text()

    def _update_align_xh_angle_change_by_delta(self, delta):
        """Update the alignment crosshair angles when the rotation is changed by a button press."""
        self.rotation_selection.setValue(self.rotation_selection.value() + delta)

    def _update_align_shape(self):
        """Update the shape of the alignment aid."""
        if self.alignment_aid.currentText() == "None":
            self.alignment_aid_shape.setVisible(False)
        else:
            self.alignment_aid_shape.setVisible(True)
            shape = self.alignment_aid.currentText()
            size = self.alignment_aid_size.value() * self.alignment_aid_scaling
            angle = self.rotation_selection.value()
            center = QtCore.QPointF(
                self.cursor_positions_selection[2].value(),
                self.cursor_positions_selection[0].value(),
            )
            path = self._create_alignment_shape(shape, center, size, angle)
            self.alignment_aid_shape.setPath(path)

    def _update_slices(self, dim_no):
        """Update the data slices when the crosshair is moved."""
        c_levels = self.colorbar.levels()  # Get starting colorbar scale levels
        self._set_slice(dim_no)
        image_item = self.image_items[dim_no]
        image_item.setImage(self.images[dim_no].values)
        if self.cscale_lock.isChecked():
            self.colorbar.setLevels(c_levels)  # Set scale to original

    def _update_DC(self, dim_no):
        """Update the DC plots when the crosshair is moved."""
        if dim_no == 1 or self.mirror_checkbox.isChecked():  # Update both DCs
            DCs_to_update = [0, 1]
        elif dim_no == 0:  # Update only DC1
            DCs_to_update = [1]
        elif dim_no == 2:  # Update only DC0
            DCs_to_update = [0]

        for DC_no in DCs_to_update:
            plot = self.DC_plot_items[DC_no][0]

            # Get the DC
            dim_no = [0, 2][DC_no]  # Select the active dim for the relevant xh plot
            DC = self._select_DC(dim_no)

            # Update the plot - get ordering correct
            if DC_no == 0:
                a, b = DC.data, DC.coords[self.dims[dim_no]].values
            else:
                a, b = DC.coords[self.dims[dim_no]].values, DC.data

            plot.setData(a, b)

            # Check if a mirrored DC exists and needs updating
            if not len(self.DC_plot_items[f"{DC_no}_m"]) == 0:
                plot_m = self.DC_plot_items[f"{DC_no}_m"][0]
                mirror_DC = sym(
                    DC,
                    flipped=True,
                    **{
                        self.dims[dim_no]: self.cursor_positions_selection[
                            dim_no
                        ].value()
                    },
                )
                if DC_no == 0:
                    a, b = mirror_DC.data, mirror_DC.coords[self.dims[dim_no]].values
                else:
                    a, b = mirror_DC.coords[self.dims[dim_no]].values, mirror_DC.data
                plot_m.setData(a, b)

    def _show_hide_mirror(self):
        """Show or hide the mirrored DCs."""
        show_mirror = self.mirror_checkbox.isChecked()
        if not show_mirror:
            for i in range(2):
                if not len(self.DC_plot_items[f"{i}_m"]) == 0:
                    self.DC_plots[i].removeItem(self.DC_plot_items[f"{i}_m"][0])
                self.DC_plot_items[f"{i}_m"] = []
        else:
            for i in range(2):
                if len(self.DC_plot_items[f"{i}_m"]) == 0:
                    self.DC_plot_items[f"{i}_m"].append(
                        self.DC_plots[i].plot(
                            [0],
                            [0],
                            pen=self.DC_mirror_pen,
                        )
                    )

        self._update_DC(1)  # Update the mirror DCs

    def _align_data(self):
        """Estimate symmetry points in the data"""
        # Get the current view range
        x_range, y_range = self.image_plots[1].getViewBox().viewRange()
        data_to_centre = self.images[1].sel(
            {self.dims[0]: slice(*y_range), self.dims[2]: slice(*x_range)}
        )

        # Get the centre of the data
        centre = estimate_sym_point(data_to_centre, dims=self.centering_dims)

        # Activate mirror mode
        self.mirror_checkbox.setChecked(True)
        self._show_hide_mirror()

        # Set main crosshair to the aligned position
        for dim, value in centre.items():
            dim_no = self.dims.index(dim)
            self.cursor_positions_selection[dim_no].setValue(value)

    def _update_cursor_stats_text(self):
        """Update the cursor stats."""
        cursor_text = ""

        # Try and parse normal emission info
        norm_emission = self._get_norm_values()
        cursor_text += _parse_norm_emission_cursor_stats(self.data, norm_emission)
        # Add metadata
        cursor_text += self.metadata_text
        # Set text
        self.cursor_stats.setText(cursor_text)

    def _get_norm_values(self):
        positions = {
            self.dims[i]: self.cursor_positions_selection[i].value() for i in range(3)
        }
        positions["azi_offset"] = -self.rotation_selection.value()
        try:
            norm_emission = self.data.metadata.get_normal_emission_from_values(positions)
        except AttributeError:
            norm_emission = {}

        return norm_emission

    def _copy_norm_values(self):
        """Copy the normal emission values to the clipboard."""
        # Current norm values
        norm_values = self._get_norm_values()
        norm_values_to_return = {
            k: str(v) if isinstance(v, pint.Quantity) else v
            for k, v in norm_values.items()
        }
        pyperclip.copy(f".metadata.set_normal_emission({norm_values_to_return})")

    # ##############################
    # Signal connections
    # ##############################
    def _connect_signals_crosshairs(self):
        # Update display boxes when crosshairs move
        for i, xh_ in enumerate(self.xhs):
            signal = xh_.xh.sigPositionChanged
            signal.connect(partial(self._update_cursor_boxes, i))  # Update cursor boxes
        # Update alignment tool rotation changes
        self.rotation_selection.valueChanged.connect(self._update_align_xh_angle)
        for angle, button in zip(
            self.rotation_delta, self.rotation_delta_buttons, strict=True
        ):
            button.clicked.connect(
                partial(self._update_align_xh_angle_change_by_delta, angle)
            )
        # Update alignment tool shape changes
        self.alignment_aid.currentIndexChanged.connect(self._update_align_shape)
        self.alignment_aid_size.valueChanged.connect(self._update_align_shape)
        # Copy current normal emission values
        self.copy_button.clicked.connect(self._copy_norm_values)

    def _connect_signals_cursor_boxes_change(self):
        # Update when cursor positions change
        for i in range(3):
            signal = self.cursor_positions_selection[i].valueChanged
            signal.connect(partial(self._update_slices, i))
            signal.connect(partial(self._update_DC, i))
            signal.connect(partial(self._update_xh_positions, i))
            signal.connect(self._update_cursor_stats_text)

    def _connect_signals_xh_span_change(self):
        # Update when span ranges changed
        for i in range(3):
            signal = self.cursor_widths_selection[i].valueChanged
            signal.connect(partial(self._update_slices, i))
            signal.connect(partial(self._update_DC, i))
            signal.connect(partial(self._update_xh_widths, i))

    def _connect_signals_align(self):
        # Connect centre data button
        self.align_button.clicked.connect(self._align_data)

    def _connect_signals_mirror_checkbox(self):
        # Connect mirror checkbox
        self.mirror_checkbox.clicked.connect(self._show_hide_mirror)

    def _connect_key_press_signals(self):
        # Connect key press signals
        self.graphics_layout.keyPressed.connect(self._key_press_event)
        self.graphics_layout.keyReleased.connect(self._key_release_event)

    def _key_press_event(self, event):
        # First deal with the modifiers
        if event.key() == self.key_modifiers["move_primary_dim"]:
            self.move_primary_dim_key_enabled = True
        elif event.key() == self.key_modifiers["hide_all"]:
            for xh in self.xhs:
                xh.set_visible(False)
        elif event.key() == self.key_modifiers["show_hide_mirror"]:
            self.mirror_checkbox.setChecked(not self.mirror_checkbox.isChecked())
            self._show_hide_mirror()
        # Now deal with the arrow keys
        elif event.key() in [
            QtCore.Qt.Key.Key_Up,
            QtCore.Qt.Key.Key_Down,
            QtCore.Qt.Key.Key_Left,
            QtCore.Qt.Key.Key_Right,
        ]:
            if self.move_primary_dim_key_enabled:
                if event.key() == QtCore.Qt.Key.Key_Up:
                    self.cursor_positions_selection[1].stepUp()
                elif event.key() == QtCore.Qt.Key.Key_Down:
                    self.cursor_positions_selection[1].stepDown()
            else:
                if event.key() == QtCore.Qt.Key.Key_Up:
                    self.cursor_positions_selection[0].stepUp()
                elif event.key() == QtCore.Qt.Key.Key_Down:
                    self.cursor_positions_selection[0].stepDown()
                elif event.key() == QtCore.Qt.Key.Key_Left:
                    self.cursor_positions_selection[2].stepDown()
                elif event.key() == QtCore.Qt.Key.Key_Right:
                    self.cursor_positions_selection[2].stepUp()

    def _key_release_event(self, event):
        if event.key() == self.key_modifiers["move_primary_dim"]:
            self.move_primary_dim_key_enabled = False
        elif event.key() == self.key_modifiers["hide_all"]:
            for xh in self.xhs:
                xh.set_visible(True)

    # ##############################
    # Helper functions
    # ##############################
    def _get_active_dim_nos(self, i):
        return [dim_no for dim_no in range(3) if dim_no != i]

    def _create_alignment_shape(self, shape, center, size, angle):
        """Create a QPainterPath for the alignment aid."""
        path = QtGui.QPainterPath()

        if shape == "Square":
            points = [
                QtCore.QPointF(
                    center.x() + size * np.cos(np.radians(angle + i * 90 + 45)),
                    center.y() + size * np.sin(np.radians(angle + i * 90 + 45)),
                )
                for i in range(4)
            ]
            path.moveTo(points[0])
            for point in points[1:]:
                path.lineTo(point)
            path.lineTo(points[0])
        elif shape == "Hexagon" or shape == "Hexagon (r30)":
            if shape == "Hexagon (r30)":
                angle += 30
            points = [
                QtCore.QPointF(
                    center.x() + size * np.cos(np.radians(angle + i * 60)),
                    center.y() + size * np.sin(np.radians(angle + i * 60)),
                )
                for i in range(6)
            ]
            path.moveTo(points[0])
            for point in points[1:]:
                path.lineTo(point)
            path.lineTo(points[0])

        elif shape == "Circle":
            rect = QtCore.QRectF(
                center.x() - size, center.y() - size, 2 * size, 2 * size
            )
            path.moveTo(center.x() + size, center.y())
            path.arcTo(rect, 0, 360)

        return path



from typing import Optional

from pydantic import create_model

from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.metadata.base_metadata_models import AxisMetadataModelWithReference


class BaseManipulatorDataLoader(BaseDataLoader):
    """Base class for data loaders for systems with manipulators.

    Notes
    -----
    THIS NEEDS UPDATING...
    Subclasses should define the `_load_manipulator_metadata` method to return a dictionary of relevant axis
    values with keys of the form `manipulator_axis` where `axis` is the names in the `_manipulator_axes` list,
    i.e. is given in :class:`peaks` convention. This method should return values as :class:`pint.Quantity` objects
    where possible to ensure units are appropriately captured and propagated. Alternatively, the main `_load_metadata`
    method can be overwritten to return the full metadata dictionary, including manipulator metadata.

    In general, subclasses will always include the 6 base axes, with a name of `None` if an axis cannot be moved.
    Then the reference positions of that axis can still be captured. Subclasses should add any additional axes
    desired via the `_add_manipulator_axes` class variable, providing a list of additional axes. Subclasses should
    also add `_update_manipulator_sign_conventions` and `_update_manipulator_name_conventions` dictionaries to update
    any sign conventions and name conventions from the default values (all axes positive and named `None`).

    `_manipulator_sign_conventions` should be a dictionary mapping the sign conventions required to get from the raw
    dimension values to standard conventions used for `peaks`. The `peaks` convention is that values of dimensions of
    the data are left matching the same sign as the experiment to make comparison with the live data more obvious,
    with any sign conversions required for data processing (e.g. $$k$$-conversion) happening under the hood. Default
    behaviour is that all axes are positive.

    `_manipulator_name_conventions` should be a dictionary mapping the `peaks` axis names to physical axis names
    on the manipulator. Default behaviour is that all axes are named as `None`, and so each physical axis should be
    added here.

    See Also
    --------
    BaseDataLoader
    BaseDataLoader._load_metadata
    """

    # Define class variables
    _loc_name = "Default Manipulator"
    _manipulator_axes = ["polar", "tilt", "azi", "x1", "x2", "x3"]
    _desired_dim_order = ["x3", "x2", "x1", "polar", "tilt", "azi"]
    _manipulator_sign_conventions = {}  # Mapping of axes to sign conventions
    _manipulator_name_conventions = {}  # Mapping of peaks axes to local names
    _manipulator_exclude_from_metadata_warn = []  # List of axes to ignore if the metadata is missing
    _metadata_parsers = [
        "_parse_manipulator_metadata"
    ]  # Specific metadata parsers to apply

    # Properties to access class variables
    @property
    def manipulator_axes(self):
        """Return the manipulator axes."""
        return self._manipulator_axes

    @property
    def manipulator_sign_conventions(self):
        """Return the manipulator sign conventions to map to `peaks` convention."""
        return self._manipulator_sign_conventions

    @property
    def manipulator_name_conventions(self):
        """Return the `peaks` --> physical manipulator name mapping."""
        return self._manipulator_name_conventions

    @classmethod
    def _parse_manipulator_metadata(cls, metadata_dict):
        """Parse metadata specific to the manipulator."""

        # Build manipulator metadata model
        fields = {
            axis: (Optional[AxisMetadataModelWithReference], None)
            for axis in cls._manipulator_axes
        }
        ManipulatorMetadataModel = create_model("ManipulatorMetadataModel", **fields)

        # Extract the relevant metadata and parse in a form for passing to the model
        manipulator_metadata_dict = {}
        for axis in cls._manipulator_axes:
            manipulator_metadata_dict[axis] = {
                "local_name": cls._manipulator_name_conventions.get(axis, None),
                "value": metadata_dict.get(f"manipulator_{axis}"),
                "reference_value": None,
            }
        # Populate the metadata model
        manipulator_metadata = ManipulatorMetadataModel(**manipulator_metadata_dict)

        # Parse list of axes that are names, so where a metadata warning should be given unless excluded by passing
        # in the class variable _manipulator_ignore_missing_metadata
        metadata_to_warn_if_missing = [
            f"manipulator_{axis}"
            for axis, name in cls._manipulator_name_conventions.items()
            if name and axis not in cls._manipulator_exclude_from_metadata_warn
        ]

        # Return the model, and a list of any metadata that should be warned if missing
        return {"_manipulator": manipulator_metadata}, metadata_to_warn_if_missing

    @classmethod
    def _parse_manipulator_references(cls, da, specified_values):
        """Methods to parse the reference values for the manipulator based on input from the user."""
        raise NotImplementedError("Subclasses should implement this method.")



"""Functions to save data as a netCDF file."""

import copy
import json
import os

import numpy as np
import pint_xarray  # noqa: F401
import xarray as xr


def _serialise_attrs(attrs):
    """Serialises attributes of a :class:`xarray.DataArray` or similar for saving in
    NetCDF/Zarr files.

    Parameters
    ----------
    attrs : dict
        The attributes dictionary of the :class:`xarray.DataArray` to be saved.

    Returns
    -------
    dict
        The serialised attributes dictionary.
    """

    # Keep track of the metadata models for de-serialisation on loading again
    metadata_models = {}

    # Make data attributes serialisable
    for attr_name, attr in attrs.copy().items():
        try:  # Attrs should generally define json method to convert to/from json string
            attrs[attr_name] = attr.json(by_alias=True)
            metadata_models[attr_name] = (
                f"{attr.__class__.__module__}.{attr.__class__.__name__}"
            )

        except AttributeError:
            # Other type serialisations
            if not isinstance(attr, np.ndarray):
                try:  # Try to convert to a json string
                    attrs[attr_name] = json.dumps(attr)
                    metadata_models[attr_name] = "json"
                except TypeError:  # Fall back to string representation
                    attrs[attr_name] = str(attr)
                    metadata_models[attr_name] = None
    attrs["metadata_models"] = json.dumps(metadata_models)

    return attrs


def _enforce_extension(fpath, required_extension):
    """Ensure that the file path has the required extension, and add it as a default if
    no extension passed.

    Parameters
    ----------
    fpath : str
        The file path to be checked.

    required_extension : str
        The required extension for the file path.

    Returns
    -------
    str : The file path with the required extension.

    Raises
    ------
    ValueError : If the file path has an incorrect extension.
    """

    f_ext = os.path.splitext(fpath)[1]
    if not f_ext:
        return f"{fpath}{required_extension}"
    elif f_ext != required_extension:
        raise ValueError(f"File path must have a {required_extension} extension.")
    else:
        return fpath


def _save_da(data, fpath):
    """Save a :class:`xarray.DataArray` or :class:`xarray.Dataset` as a NetCDF file.

    Parameters
    ----------
    data : xarray.DataArray or xarray.DataSet
        The data to be saved.

    fpath : str
        The path to the file to be created.
    """
    # Ensure the file path has the correct extension
    fpath = _enforce_extension(fpath, ".nc")

    # Copy the original attributes of the data to reset them after saving
    original_attrs = copy.deepcopy(data.attrs)

    # Add a history entry for the data saving to the analysis_history
    data.history.add(f"Data saved as a NetCDF file to {fpath}.")

    # Prepare the data to be saved by serialising the data attributes
    data.attrs.update(_serialise_attrs(data.attrs))

    # Save the data and reset the attributes to their original state
    data.pint.dequantify().to_netcdf(fpath)
    data.attrs = original_attrs


def _save_dt(data, fpath):
    """Save a :class:`xarray.DataTree` as a zarr file.

    Parameters
    ----------
    data : xarray.DataTree
        The data to be saved.

    fpath : str
        The path to the file to be created.
    """
    # Ensure the file path has the correct extension
    fpath = _enforce_extension(fpath, ".zarr")

    def _quantify_da_in_dt(data):
        return data.pint.quantify()

    def _add_history_entry(da):
        da.history.add(f"Data saved as part of a DataTree Zarr file to {fpath}.")
        return da

    def _serialise_da_in_dt(data):
        """Serialises the attributes of a :class:`xarray.DataArray` or similar for saving
        in NetCDF/Zarr files."""
        data.attrs.update(_serialise_attrs(data.attrs))
        return data

    def _parse_nodes_in_dt(ds):
        """Parse the attributes of the current datatree node

        Parameters
        ----------
        ds : xarray.DataArray or xarray.Dataset
            The data to be parsed

        Returns
        -------
        xarray.Dataset : The parsed data
        """
        # Add history entries to the Dataset or DataArray as appropriate
        if "_analysis_history" in ds.attrs:
            ds = _add_history_entry(ds)
        else:
            ds = ds.map(_add_history_entry)
        # Serialise attributes of each DataArray, mapping over the DataTree & Datasets
        ds = _serialise_da_in_dt(ds).map(_serialise_da_in_dt)

        # Dequantify the Dataset
        ds = ds.pint.dequantify()
        return ds

    # Iterate through the DataTree and store the original attributes of each node
    original_attrs = []
    for node in data.subtree:
        if node.attrs:
            original_attrs.append(copy.deepcopy(node.attrs))
        else:
            original_attrs.append({})

    # Map over the DataTree to serialise the attributes and dequantify the DataArrays
    data = data.map_over_datasets(_parse_nodes_in_dt)

    # Save data as a zarr file
    data.to_zarr(fpath)

    # Reset the attributes of the data to their original state
    for node, attrs in zip(data.subtree, original_attrs, strict=True):
        node.attrs = attrs

    # Re-quanitfy the data
    data = data.map_over_datasets(_quantify_da_in_dt)


def save(data, fpath):
    """This function saves data in the :class:`xarray.DataArray` or
    :class:`xarray.DataSet` format as a NetCDF file or a :class:`xarray.DataTree` as a
    zarr file. These formats are restricted in what types of attributes can be saved.
    This function attempts to save peaks metadata attributes in a way that can be parsed
    again when reloading but if custom attributes are used or the file is not saved and
    loaded with the same version of `peaks`, there may be some loss of metadata.

    Parameters
    ------------
    data : xarray.DataArray or xarray.DataSet or xarray.DataTree
        The data to be saved

    fpath : str
        Path to the file to be created. Note: the extension of the file must either be omitted, or must be specified
        as .nc for a :class:`xarray.DataArray` or :class:`xarray.DataSet` and .zarr for a :class:`xarray.DataTree`.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # Load some data and do some processing
        data = pks.load('my_file.ibw').k_convert()

        # Save the data as a NetCDF file
        pks.save(data, 'my_file.nc')

    """

    if isinstance(data, (xr.DataArray, xr.Dataset)):
        _save_da(data, fpath)
    elif isinstance(data, xr.DataTree):
        return _save_dt(data, fpath)



"""General data loading helper functions."""

# Phil King 15/05/2021
# Brendan Edwards 26/02/2024

import numpy as np

from peaks.core.fileIO.data_loading import _extract_mapping_metadata, _make_DataArray
from peaks.core.utils.misc import analysis_warning


def make_hv_scan(data):
    """Function to combine multiple dispersions (measured at different hv values) into a single hv scan DataArray.

    Parameters
    ------------
    data : list
        Any number of :class:`xarray.DataArray` dispersions (measured at different hv values) to combine into an hv scan.

    Returns
    ------------
    hv_scan : xarray.DataArray
        The resulting hv scan DataArray.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp_70eV = load('disp70eV.ibw')
        disp_72eV = load('disp72eV.ibw')
        disp_74eV = load('disp74eV.ibw')
        disp_76eV = load('disp76eV.ibw')
        disp_78eV = load('disp78eV.ibw')
        disp_80eV = load('disp80eV.ibw')

        # Combine dispersions (measured at different hv values) into a single hv scan DataArray
        hv_scan = make_hv_scan([disp_70eV, disp_72eV, disp_74eV, disp_76eV, disp_78eV, disp_80eV])


    """

    # Define the new scan type
    scan_type = "hv scan"

    # Ensure the dispersions are arranged in order of increasing hv
    hvs_and_disps = []
    for disp in data:
        hvs_and_disps.append([float(disp.hv), disp])
    hvs_and_disps.sort()

    # Extract the data spectrum and the hv, KE and theta_par coordinates
    spectrum = []
    hv_values = []
    KE_values = []
    for hv, disp in hvs_and_disps:
        # Want theta_par as the first dimension to get the expected spectrum shape in the _make_DataArray function
        if disp.dims[0] == "eV":
            spectrum.append(disp.data.T)
        else:
            spectrum.append(disp.data)
        hv_values.append(hv)
        KE_values.append(disp.coords["eV"].data)
    theta_par_values = data[0].theta_par.data

    # We want to save the kinetic energy coordinates of the first scan, and also the corresponding offsets for
    # successive scans (KE_delta)
    KE0 = np.array(KE_values)[
        :, 0
    ]  # Get  the maximum kinetic energy of the scan as a function of photon energy
    KE_delta = KE0 - KE0[0]  # Get the change in KE value of detector as a function of hv

    # Define dictionary to be sent to the _make_DataArray function to make a xarray.DataArray
    data_dict = {
        "scan_type": scan_type,
        "spectrum": np.array(spectrum),
        "hv": hv_values,
        "theta_par": theta_par_values,
        "eV": KE_values[0],
        "KE_delta": KE_delta,
    }

    # Make an hv scan DataArray
    hv_scan = _make_DataArray(data_dict)

    # Ensure that the dimensions of the hv scan are arranged in the standard order
    hv_scan = hv_scan.transpose(
        "hv", "eV", "defl_par", "theta_par", "k_par", missing_dims="ignore"
    )

    # Add metadata to the new hv scan DataArray
    hv_scan.attrs = data[0].attrs
    hv_scan.attrs["scan_name"] = "Manual hv_scan"
    hv_scan.attrs["scan_type"] = "hv scan"
    hv_scan.attrs["hv"] = _extract_mapping_metadata(hv_values, num_dp=2)

    # Update analysis history
    hv_scan.history.add("Combined multiples dispersions into an hv scan")

    # Display warning explaining how kinetic energy values are saved
    warn_str = (
        "The kinetic energy coordinates saved are that of the first scan. The corresponding offsets "
        "for successive scans are included in the KE_delta coordinate. Run DataArray.disp_from_hv(hv), "
        "where DataArray is the loaded hv scan xarray.DataArray and hv is the relevant photon energy, "
        "to extract a dispersion at using the proper kinetic energy scaling for that photon energy."
    )
    analysis_warning(warn_str, title="Loading info", warn_type="info")

    return hv_scan



from xarray import open_datatree

from peaks.core.fileIO.loaders.netcdf import NetCDFLoader
from peaks.core.fileIO.loc_registry import register_loader
from peaks.core.utils.misc import analysis_warning


@register_loader
class ZarrLoader(NetCDFLoader):
    """Class for re-opening Zarr files saved from :class:`peaks.core.fileIO.data_saving.save`.

    Notes
    -----
    If this is a Zarr file, it should be a DataTree
    """

    _loc_name = "Zarr"

    @classmethod
    def _load(cls, fpath, lazy, metadata, quiet):
        """Load the data from a Zarr file.

        Parameters
        ----------
        fpath : str
            Path to the file to be loaded.
        lazy : bool
            Whether to load the data lazily.
        metadata : bool
            Whether to attempt to parse the metadata into full `peaks` format.
        quiet : bool
            Whether to suppress print output.

        Returns
        -------
        data : xarray.DataArray or xarray.DataSet
            The loaded data.
        """

        # Load Zarr file as xarray.DataArray or xarray.Dataset
        data = open_datatree(fpath, engine="zarr", chunks={})

        # Parse the metadata if requested
        if metadata:
            data = data.map_over_datasets(ZarrLoader._parse_ds_metadata)

        # Quantify the data
        data = data.map_over_datasets(ZarrLoader._quantify_da_in_dt)

        # Actually load the data
        if not lazy:
            data = data.map_over_datasets(ZarrLoader._load_all)
        elif not quiet:
            analysis_warning(
                "The data is lazily loaded by default for loading from a Zarr store. "
                "Use the .compute() method on the individual data entries to load each into memory. "
                "To eagerly load the entire DataTree, pass the `lazy=False` argument to the load "
                "function. Ensure that the total data contents is not too large to fit in memory.",
                title="Loading info",
                warn_type="info",
            )

        return data

    @staticmethod
    def _parse_ds_metadata(ds):
        return ZarrLoader._parse_da_metadata(ds).map(ZarrLoader._parse_da_metadata)

    @staticmethod
    def _parse_da_metadata(data):
        ZarrLoader._parse_metadata(data)
        return data

    @staticmethod
    def _quantify_da_in_dt(data):
        return data.pint.quantify()

    @staticmethod
    def _load_all(data):
        return data.compute()



"""Functions to load data into DataArray format."""

import glob
import os

import dask
import pint
from tqdm.notebook import tqdm

from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.options import opts
from peaks.core.utils.datatree_utils import _dataarrays_to_datatree
from peaks.core.utils.misc import analysis_warning


def load(
    fpath,
    lazy=None,
    loc=None,
    metadata=True,
    parallel=False,
    names=None,
    quiet=False,
    **kwargs,
):
    """Core function to load data.

    Parameters
    ------------
    fpath : str, list
        Either the full file path(s), or the remainder of the file name(s) not already
        specified in the file global options (see Notes).

    lazy : str, bool, optional
        Whether to load data in a lazily evaluated dask format. Set explicitly using
        True/False Boolean. Defaults to `None` where a file is only loaded in the dask
        format if its spectrum is above threshold set in `opts.FileIO.lazy_size` or is a
        Zarr store, where it is lazily loaded by default.

    loc : str, optional
        Location identifier for where data was acquired. Defaults to `None`, where the
        location will be attempted to be automatically determined, unless a value is
        defined in `opts.FileIO.loc`. If `loc` is specified in the function call, this
        takes priority over the value in `opts.FileIO.loc`.

    metadata : bool, optional
        Whether to attempt to load metadata into the attributes of the
        :class:`xarray.DataArray`. Defaults to True.

    parallel : bool, optional
        Whether to load data in parallel when multiple files are being loaded. Only
        compatible with certain file types such as those based on the h5py format, e.g.
        .nxs files. Takes priority over lazy, enforcing that all data is computed and
        loaded into memory. Defaults to False.

    names : list, optional
        List of names to assign to the branches of the :class:xarray.DataTree when
        mutliple scans loaded simultaneously. If provided, should be a list of unique
        strings of the same length as the number of scans being loaded. If not, or if
        not provided, the names will be the file names. Defaults to None.

    quiet : bool, optional
        Whether to suppress analysis warnings when loading data. Defaults to False.

    kwargs : dict
        Additional keyword arguments to pass to the data loader.

    Returns
    ------------
    loaded_data : xarray.DataArray, xarray.DataSet, xarray.DataTree
        The loaded data.

    Notes
    ------------
    Much of file path can be set by :class:`peaks.core.options.FileIO` global options:

        opts.FileIO.path : str, list
            Path (or list of paths) to folder(s) where data is stored,
            e.g. `opts.FileIO.path = 'C:/User/Documents/i05-1-123'`.

        opts.FileIO.ext : str, list
            Extension (or list of) of data, e.g. `opts.FileIO.ext = ['ibw', 'zip']`.

        opts.FileIO.loc : str
            Location identifier for where data was acquired, e.g. `opts.FileIO.loc = 'MAX IV Bloch'`.
            Current supported options can be obtained using `pks.locs()`.

        opts.FileIO.lazy_size : int
            Size in bytes above which data is loaded in a lazily evaluated dask format.
            Defaults to 1 GB.

    Examples
    ------------
    Example usage is as follows::

        import peaks as pks

        # Load data using a complete data path
        disp1 = pks.load('C:/User/Documents/disp1.ibw')
        FM1 = pks.load('C:/User/Documents/FM1.ibw')

        # Define global options to define file loading
        pks.opts.FileIO.set(path = 'C:/User/Documents/Data/',
                            ext = ['ibw', 'zip'])

        # Load data without needing to define data path or extension
        disp2 = pks.load('disp2')
        FM2 = pks.load('FM2')

        # Define global options to define file structure, including part of the scan name
        pks.opts.FileIO.path = 'C:/User/Documents/Data/i05-1-123'
        pks.opts.FileIO.ext = 'nxs'

        # Load data without needing to define data path or extension, nor the repeated
        # part of the scan name
        disp3 = load(456)
        disp4 = load(457)

        # Provide part of the scan name using a glob character
        disps = load('45*')

        # Load multiple files at once
        disps = load([456, 457, 458])

        # Still can load data using a complete data path
        # global options defined in `pks.opts.FileIO` will be ignored
        disp1 = load('C:/User/Documents/Data/disp1.ibw')

        # Load data in a lazily evaluated dask format
        disp1 = load('C:/User/Documents/Data/disp1.ibw', lazy=True)

        # Load data without metadata
        disp1 = load('C:/User/Documents/Data/disp1.ibw', metadata=False)

        # Load data file for a defined location (use if automatic location ID fails)
        disp1 = load('C:/User/Documents/Data/disp1.ibw', loc='MAX IV Bloch')

        # Alternatively could define location using global options.
        # Here loc will be defined as 'MAX IV Bloch'
        pks.opts.FileIO.loc = 'MAXIV_Bloch_A'
        disp1 = load('C:/User/Documents/Data/disp1.ibw')

        # Set the data size to trigger lazy loading by default to 500 MB
        pks.opts.FileIO.lazy_size = 500000000
        FM3 = pks.load('C:/User/Documents/FM3.ibw')

        # Can also use with a context manager to temporarily set the global options
        with pks.opts as opts:
            opts.FileIO.path = 'C:/User/Documents/Data/'
            opts.FileIO.ext = ['ibw', 'zip']
            opts.FileIO.loc = 'MAXIV_Bloch_A'
            opts.FileIO.lazy_size = 500000000

            # Load data without needing to define data path or extension
            disp2 = pks.load('disp2')
            FM2 = pks.load('FM2')
    """

    load_opts = {
        "lazy": lazy,
        "loc": loc,
        "metadata": metadata,
        "parallel": parallel,
        "names": names,
        "quiet": quiet,
    }
    load_opts.update(kwargs)

    # If a full file path provided, always load only this
    if isinstance(fpath, str) and os.path.exists(fpath) and "." in fpath:
        return _load_data(fpath, **load_opts)

    # Set placeholder file path and extension
    base_path = [None]
    ext = [""]

    # If FileIO.path is defined, make a list of the inputted path(s)
    if opts.FileIO.path:
        base_path = (
            [str(path) for path in opts.FileIO.path]
            if isinstance(opts.FileIO.path, list)
            else [str(opts.FileIO.path)]
        )

    # If file.ext is defined, make a list of the inputted extension(s)
    if opts.FileIO.ext:
        ext = [opts.FileIO.ext] if isinstance(opts.FileIO.ext, str) else opts.FileIO.ext

    # If the parameter loc is not defined and opts.FileIO.loc is defined, update loc
    if loc is None:
        load_opts["loc"] = opts.FileIO.loc

    # Ensure that fpath is a list of strings
    fpath = [str(path) for path in fpath] if isinstance(fpath, list) else [str(fpath)]

    # Obtain all possible file addresses
    possible_file_addresses = []
    # Loop through file names
    for file_name in fpath:
        # Remove any extensions from fname, adding them to ext
        file_name, fname_ext = os.path.splitext(file_name)
        if fname_ext != "" and fname_ext[1:] not in ext:
            ext.append(fname_ext[1:])

        # Add raw address file_name to valid_file_addresses
        possible_file_addresses.append(file_name)

        # Loop through base file paths and add file_path/file_name combinations
        for file_path in base_path:
            if file_path:
                # Check if file_path is already an existing folder
                if os.path.exists(file_path) and os.path.isdir(file_path):
                    # If so, join properly so it doesn't matter if a terminator was given
                    possible_file_addresses.append(os.path.join(file_path, file_name))
                else:
                    # Otherwise, must have had a partial file name, so just append
                    possible_file_addresses.append(file_path + file_name)

    # Obtain all valid file addresses
    file_list = []
    for address in possible_file_addresses:
        for extension in ext:
            # Construct the path pattern
            if extension:
                pattern = f"{address}.{extension}"
            else:
                pattern = address

            # Use glob to expand wildcards
            matched_files = glob.glob(pattern)

            # Filter for existing paths and add them
            for file in matched_files:
                if os.path.exists(file):
                    file_list.append(file)

    # Remove duplicates from file_list
    file_list = list(set(file_list))

    # Load data by calling load_data if a valid file has been found. If not raise an error
    if len(file_list) > 0:
        return _load_data(file_list, **load_opts, **kwargs)

    raise Exception("No valid file paths could be found.")


def _load_data(fpath, lazy, loc, metadata, parallel, names, quiet, **kwargs):
    """Function to handle loading of single or multiple data files into the xarray DataArray format.

    Returns
    ------------
    loaded_data : xarray.DataArray, xarray.DataSet, list
        The loaded data.

    See Also
    ------------
    load : Public function to load data, which sets much of the fpath and defines the options to pass to `_load_data`.

    """

    load_opts = {"lazy": lazy, "loc": loc, "metadata": metadata, "quiet": quiet}
    load_opts.update(kwargs)

    # Ensure fname is of type list
    if not isinstance(fpath, list):
        fpath = [fpath]

    # Define an empty list to store loaded data
    loaded_data = []

    # If the files have been requested to be loaded in parallel
    if parallel:
        # Loop through and set up the dask delayed function which will facilitate data loading in parallel
        for single_fpath in fpath:
            loaded_data.append(
                dask.delayed(BaseDataLoader.load(fpath=single_fpath, **load_opts))
            )

        # Perform the data loading in parallel
        loaded_data = dask.compute(*loaded_data)

        # If Lazy is not False, display message informing the user that setting parallel to True means all data is
        # loaded into memory and cannot be lazily evaluated
        if lazy:
            analysis_warning(
                "By setting parallel=True, the data has been computed and loaded into memory. This means that the "
                "data is unable to be lazily evaluated in the dask format. If a lazy evaluation is required, set "
                "parallel=False",
                title="Loading info",
                warn_type="danger",
            )

    # If not, load files sequentially
    else:
        for single_fpath in tqdm(
            fpath,
            desc="Loading data",
            disable=len(fpath) == 1,
        ):
            loaded_data.append(BaseDataLoader.load(fpath=single_fpath, **load_opts))

    # Check if any of the loaded data have been lazily evaluated. If so, inform the user
    for data in loaded_data:
        try:
            if isinstance(data.data, dask.array.core.Array) or (
                isinstance(data.data, pint.Quantity)
                and isinstance(data.data.magnitude, dask.array.core.Array)
            ):
                analysis_warning(
                    "DataArray has been lazily evaluated in the dask format (set lazy=False to load as DataArray in xarray "
                    "format). Use the .compute() method to load DataArray into RAM in the xarray format, or the .persist() "
                    "method to instead load DataArray into RAM in the dask format. Note: these operations load all of the "
                    "data into memory, so large files may require an initial reduction in size through either a slicing or "
                    "binning operation.",
                    title="Loading info",
                    warn_type="info",
                )
                break
        except AttributeError:
            pass

    # If there is only one loaded item in loaded_data, return the xr.DataArray (xr.DataSet) entry instead of a list
    if len(loaded_data) == 1:
        return loaded_data[0]

    # Otherwise, parse these into a DataTree
    return _dataarrays_to_datatree(loaded_data, names)



"""Functions to load data from the Casiopee beamline at Soleil."""

import itertools
import os
from os.path import isfile, join

import natsort
import numpy as np
import pint_xarray
from tqdm.notebook import tqdm

from peaks.core.fileIO.base_arpes_data_classes.base_ses_class import SESDataLoader
from peaks.core.fileIO.loc_registry import register_loader

ureg = pint_xarray.unit_registry


@register_loader
class CASIOPEEArpesLoader(SESDataLoader):
    """Data loader for Casiopee ARPES system at Soleil.

    Notes
    ------------
    Not thoroughly tested yet

    """

    _loc_name = "Soleil_Casiopee_ARPES"
    _loc_description = "ARPES branch of Casiopee beamline at Soleil"
    _loc_url = "https://www.synchrotron-soleil.fr/en/beamlines/cassiopee"
    _analyser_slit_angle = 0 * ureg("deg")

    _manipulator_name_conventions = {
        "polar": "theta",
        "tilt": "tilt",
        "azi": "phi",
        "x1": "x",
        "x2": "z",
        "x3": "y",
    }
    _SES_metadata_units = {
        f"manipulator_{dim}": ("mm" if dim in ["x1", "x2", "x3"] else "deg")
        for dim in _manipulator_name_conventions.keys()
    }

    @classmethod
    def _load_data(cls, fpath, lazy):
        """Load SOLEIL Casiopee data."""

        if os.path.splitext(fpath)[1] in [".txt", ".zip", ".ibw"]:
            # Then should be standard SES format
            return super()._load_data(fpath, lazy)

        # Otherwise, need to use custom SOLEIL Casiopee loader functions
        # Load and cache the metadata
        metadata_dict_ses_keys = cls._load_metadata(fpath, return_in_SES_format=True)
        # Cahce it in the metadata cache to avoid having to load it again later
        cls._metadata_cache[fpath] = metadata_dict_ses_keys

        # Extract the relevant filenames into two lists
        file_list_ROI = natsort.natsorted(
            [
                item
                for item in os.listdir(fpath)
                if "ROI1" in item and isfile(join(fpath, item)) and item.endswith(".txt")
            ]
        )
        file_list_i = natsort.natsorted(
            [
                item
                for item in os.listdir(fpath)
                if "_i" in item and isfile(join(fpath, item))
            ]
        )

        # Extract kinetic energy and theta_par values
        eV_values = np.fromstring(metadata_dict_ses_keys["Dimension 1 scale"], sep=" ")
        # Precision in the file isn't good enough, so do a linear scaling between end points
        eV_values = np.linspace(eV_values[0], eV_values[-1], len(eV_values))
        eV_units = cls._parse_SES_units_from_name(
            metadata_dict_ses_keys.get("Dimension 1 name", "")
        )
        theta_par_values = np.fromstring(
            metadata_dict_ses_keys["Dimension 2 scale"], sep=" "
        )
        theta_par_units = cls._parse_SES_units_from_name(
            metadata_dict_ses_keys.get("Dimension 2 name", "")
        )
        theta_par_label = "theta_par" if theta_par_units == "deg" else "y_scale"

        # Loop through images and extract potential mapping coordinates polar and hv (one of polar or hv will vary, one
        # will be static)
        polar_values = []
        hv_values = []

        def _SES_find(lines, item):
            # Loop over lines to extract the line starting with the desired keyword.
            for line in lines:
                if line.startswith(item) and "=" in line:
                    line_contents = line.split("=")[-1].strip()
                    break
                elif line.startswith(item) and ":" in line:
                    line_contents = line.split(":")[-1].strip()
                    break

            return line_contents

        for file_i in file_list_i:
            with open(fpath + "/" + file_i) as f:
                lines = f.readlines()
                polar = float(_SES_find(lines, "theta (deg)"))
                hv = float(_SES_find(lines, "hv (eV)"))
            polar_values.append(polar)
            hv_values.append(hv)

        polar_values = np.array(polar_values)
        hv_values = np.array(hv_values)

        if abs(hv_values[-1] - hv_values[0]) > 1:  # Should be an hv scan
            dims = ["hv", theta_par_label, "eV"]
            coords = {
                "hv": hv_values,
                theta_par_label: theta_par_values,
                "eV": eV_values,
            }
            units = {
                "hv": "eV",
                theta_par_label: theta_par_units,
                "eV": eV_units,
                "spectrum": "counts",
            }
            spectrum = np.zeros((len(hv_values), len(theta_par_values), len(eV_values)))
        else:
            dims = ["polar", theta_par_label, "eV"]
            coords = {
                "polar": polar_values,
                theta_par_label: theta_par_values,
                "eV": eV_values,
            }
            units = {
                "polar": "deg",
                theta_par_label: theta_par_units,
                "eV": eV_units,
                "spectrum": "counts",
            }
            spectrum = np.zeros(
                (len(polar_values), len(theta_par_values), len(eV_values))
            )

        # Define the shape of the individual 2D data slices of the 3D data to be extracted
        slice_shape = (len(theta_par_values), len(eV_values))

        # Loop through the individual 2D data slices of the 3D data and extract the spectrum
        for i, file in tqdm(
            enumerate(file_list_ROI),
            total=len(file_list_ROI),
            desc="Loading data slices",
        ):
            spectrum[i, :, :] = cls._load_CASSIOPEE_slice(fpath, file, slice_shape)

        # If scan is hv scan, add the KE_delta coord
        if "hv" in dims:
            KE_delta = hv_values - hv_values[0]
            coords["KE_delta"] = ("hv", KE_delta)

        return {"spectrum": spectrum, "coords": coords, "dims": dims, "units": units}

    @staticmethod
    def _load_CASSIOPEE_slice(folder, file, slice_shape):
        """This function loads a single 2D slice of 3D data (either a Fermi map or hv scan) that was obtained at the
        CASSIOPEE beamline at SOLEIL.

        Parameters
        ------------
        folder : str
            Path to the folder of the 3D data that a data slice will be loaded from.

        file : str
            Name of the file within the folder of the 3D data corresponding to the data slice to be loaded.

        slice_shape : tuple
            Shape of the numpy.ndarray data slice to be extracted.

        Returns
        ------------
        slice_data : numpy.ndarray
            A single 2D data slice of the 3D data stored in folder.

        Examples
        ------------
        Example usage is as follows::

            from peaks.core.fileIO.loaders.CASSIOPEE import _load_CASSIOPEE_slice

            folder = 'C:/User/Documents/Research/FS1'

            # Extract the sixth slice of data from the FS1 file obtained at the CASSIOPEE beamline at SOLEIL
            data_slice = _load_CASSIOPEE_slice(folder, 'FS1_1_ROI6_.txt')

        """

        # Open the file within the folder of the 3D data corresponding to the data slice to be loaded, loop through the
        # lines, and stop when the scan data is reached
        with open(folder + "/" + file) as f:
            lines = f.readlines()
            for counter, line in enumerate(lines):  # noqa: B007
                if line.startswith("inputA="):
                    break

            # Define numpy.ndarray to store data slice
            slice_data = np.zeros(slice_shape)

            # Extract the data slice
            for i in range(len(lines) - (counter + 2)):
                slice_data[:, i] = np.fromstring(lines[counter + 2 + i], sep="\t")[1:]

            return slice_data

    @classmethod
    def _load_metadata(cls, fpath, return_in_SES_format=False):
        """Load SOLEIL Casiopee metadata."""

        if os.path.splitext(fpath)[1] in [".txt", ".zip", ".ibw"]:
            # Then should be standard SES format
            return super()._load_metadata(fpath, return_in_SES_format)

        # Otherwise, need to use custom SOLEIL Casiopee loader functions

        # Extract the relevant filenames into two lists
        file_list_ROI = natsort.natsorted(
            [
                item
                for item in os.listdir(fpath)
                if "ROI1" in item and isfile(join(fpath, item))
            ]
        )

        # Open the first file and extract the lines containing metadata
        with open(fpath + "/" + file_list_ROI[0]) as f:
            metadata_lines = [line for line in itertools.islice(f, 0, 45)]

        metadata_dict_SES_keys = cls._SES_metadata_to_dict_w_SES_keys(metadata_lines)

        # Check if there is additional run mode information in the metadata
        try:
            run_mode_info_start_index = metadata_lines.index("[Run Mode Information]")
            run_mode_info_stop_index = (
                metadata_lines[run_mode_info_start_index:].index("")
                + run_mode_info_start_index
            )
            metadata_dict_SES_keys["Run Mode Information"] = metadata_lines[
                run_mode_info_start_index + 1 : run_mode_info_stop_index
            ]

        except ValueError:
            pass

        if return_in_SES_format:
            return metadata_dict_SES_keys

        # Convert the metadata to the peaks convention and return
        return cls._SES_metadata_dict_keys_to_peaks_keys(metadata_dict_SES_keys)



"""Functions used for derivative operations on data."""

import numpy as np
from scipy.ndimage import gaussian_gradient_magnitude

from peaks.core.utils.misc import analysis_warning


def deriv(data, dims):
    """General function to perform differentiations along the specified dimensions of data.

    Parameters
    ------------
    data : xarray.DataArray
        The data to differentiate.

    dims : str, list
        Dimension(s) to perform differentiation(s) along. Use a str for a single differentiation, and a list of strs
        for multiple differentiations.

    Returns
    ------------
    deriv_data : xarray.DataArray
        The differentiated data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        # Differentiate the dispersion along eV
        disp_deriv = disp.deriv('eV')

        # Double differentiate the dispersion along eV
        disp_deriv = disp.deriv(['eV', 'eV'])

        # Differentiate the dispersion along eV and then along theta_par
        disp_deriv = disp.deriv(['eV', 'theta_par'])

        # Smooth and then differentiate the dispersion along eV
        disp_deriv = disp.smooth(eV=0.05).deriv('eV')

    """

    # Copy the input data to prevent overwriting issues
    deriv_data = data.copy(deep=True)

    # Ensure dims is of type list
    if not isinstance(dims, list):
        dims = [dims]

    # Save the attributes as these currently get killed by xarray's default differentiate function
    # Note: keep_attrs option still does not exist for differentiate in xarray version 2023.9.0
    attributes = deriv_data.attrs

    # List to store analysis history
    hist_list = []

    # Iterate through specified dimensions and perform differentiations
    for dim in dims:
        if (
            dim not in deriv_data.dims
        ):  # If supplied dimension is not a valid, raise an error
            raise Exception(
                "{dim} is not a valid dimension of the inputted DataArray.".format(
                    dim=dim
                )
            )
        deriv_data = deriv_data.differentiate(dim)  # Perform differentiation
        hist_list.append(
            "Applied differentiation along {dim}".format(dim=dim)
        )  # Update analysis history list

    # Rewrite attributes
    deriv_data.attrs = attributes

    # Update analysis history
    for hist in hist_list:
        deriv_data.history.add(hist)

    return deriv_data


def d2E(data):
    """Shortcut function to perform a double differentiation along the eV dimension of data.

    Parameters
    ------------
    data : xarray.DataArray
        The data to differentiate.

    Returns
    ------------
    deriv_data : xarray.DataArray
        The differentiated data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        # Double differentiate the dispersion along eV
        disp_deriv = disp.d2E()

        # Smooth and then double differentiate the dispersion along eV
        disp_deriv = disp.smooth(eV=0.05).d2E()

    """

    # If eV is not a valid dimension, raise an error
    if "eV" not in data.dims:
        raise Exception("eV is not a valid dimension of the inputted DataArray.")

    # Double differentiate the data along eV axis
    deriv_data = data.deriv(["eV", "eV"])

    return deriv_data


def d2k(data):
    """Shortcut function to perform a double differentiation along the momentum (or angle) dimension of data.

    Parameters
    ------------
    data : xarray.DataArray
        The data to differentiate.

    Returns
    ------------
    deriv_data : xarray.DataArray
        The differentiated data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        # Double differentiate the dispersion along the angle dimension
        disp_deriv = disp.d2k()

        # Smooth and then double differentiate the dispersion along the angle dimension
        disp_deriv = disp.smooth(theta_par=0.5).d2k()

    """

    # Work out correct variable for differentiation direction (i.e. is data in angle or k-space)
    coords = list(data.dims)
    if "eV" in coords:
        coords.remove("eV")
    coord = coords[-1]  # Should always be the last one if data loading is consistent

    # Double differentiate the data along the momentum (or angle) axis
    deriv_data = data.deriv([coord, coord])

    return deriv_data


def dEdk(data):
    """Shortcut function to perform sequential differentiations along the eV then momentum (or angle) dimensions of
    data.

    Parameters
    ------------
    data : xarray.DataArray
        The data to differentiate.

    Returns
    ------------
    deriv_data : xarray.DataArray
        The differentiated data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        # Sequentially differentiate the dispersion along the eV then angle dimension
        disp_deriv = disp.dEdk()

        # Smooth and then sequentially differentiate the dispersion along the eV then angle dimension
        disp_deriv = disp.smooth(eV=0.05, theta_par=0.5).dEdk()

    """

    # Get inputted DataArray dimensions
    coords = list(data.dims)

    # If inputted DataArray is not 2D or higher, raise an error
    if len(coords) < 2:
        raise Exception("Inputted DataArray must be at least 2D.")

    # If eV is not a valid dimension, raise an error
    if "eV" not in data.dims:
        raise Exception("eV is not a valid dimension of the inputted DataArray.")

    # Work out correct variable for the angle/momentum direction
    coords.remove("eV")
    coord = coords[-1]  # Should always be the last one if data loading is consistent

    # Sequentially differentiate the dispersion along the eV then angle/momentum dimension
    deriv_data = data.deriv(["eV", coord])

    return deriv_data


def dkdE(data):
    """Shortcut function to perform sequential differentiations along the momentum (or angle) then eV dimensions of
    data.

    Parameters
    ------------
    data : xarray.DataArray
        The data to differentiate.

    Returns
    ------------
    deriv_data : xarray.DataArray
        The differentiated data.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        # Sequentially differentiate the dispersion along the angle then eV dimension
        disp_deriv = disp.dkdE()

        # Smooth and then sequentially differentiate the dispersion along the angle then eV dimension
        disp_deriv = disp.smooth(theta_par=0.5, eV=0.05).dkdE()

    """

    # Get inputted DataArray dimensions
    coords = list(data.dims)

    # If inputted DataArray is not 2D or higher, raise an error
    if len(coords) < 2:
        raise Exception("Inputted DataArray must be at least 2D.")

    # If eV is not a valid dimension, raise an error
    if "eV" not in data.dims:
        raise Exception("eV is not a valid dimension of the inputted DataArray.")

    # Work out correct variable for the angle/momentum direction
    coords.remove("eV")
    coord = coords[-1]  # Should always be the last one if data loading is consistent

    # Sequentially differentiate the dispersion along the eV then angle/momentum dimension
    deriv_data = data.deriv([coord, "eV"])

    return deriv_data


def curvature(data, **parameter_kwargs):
    """Perform 2D curvature analysis of data (see Rev. Sci. Instrum.  82, 043712 (2011) for analysis procedure).

    Parameters
    ------------
    data : xarray.DataArray
        The data to perform curvature analysis on.

    **parameter_kwargs : float
        Curvature analysis free parameters in the format axis=value, e.g. theta_par=0.1. Free parameters must be defined
        for both axes of the data. Set a given axis free parameter to 0 to obtain 1D curvature analysis for the other
        axis.

    Returns
    ------------
    curv_data : xarray.DataArray
        The data following curvature analysis.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        # Perform curvature analysis on the dispersion using free parameters for the theta_par and eV axes of 10 and 1
        disp_curv = disp.curvature(theta_par=10, eV=1)

        # Smooth and then perform curvature analysis on the dispersion using free parameters for the
        # theta_par and eV axes of 10 and 1
        disp_curv = disp.smooth(eV=0.03, theta_par=0.3).curvature(theta_par=10, eV=1)

    """

    # Check data is 2D
    if len(data.dims) != 2:
        raise Exception("Function only acts on 2D data.")

    # Check free parameters have been provided for both axes of the data
    for dim in data.dims:
        if (
            dim not in parameter_kwargs
        ):  # Raise error if a dimension of the data is not defined in parameter_kwargs
            raise Exception(
                "Function requires free parameters to be defined for both axes of the data."
            )

    # Copy the input xarray to prevent overwriting issues
    curv_data = data.copy(deep=True)

    # Save the attributes as these get killed by the curvature analysis summation
    attributes = curv_data.attrs

    # Determine relevant axes and get associated free parameters
    dimx = curv_data.dims[0]
    dimy = curv_data.dims[1]
    Cx = parameter_kwargs[dimx]
    Cy = parameter_kwargs[dimy]

    # Determine various derivatives used in curvature analysis (0 and 1 in following notation represent dimx and dimy)
    dx = curv_data.deriv(dimx)  # d/dx
    d2x = curv_data.deriv([dimx, dimx])  # d^2/dx^2
    dy = curv_data.deriv(dimy)  # d/dy
    d2y = curv_data.deriv([dimy, dimy])  # d^2/dy^2
    dxdy = curv_data.deriv([dimx, dimy])  # d^2/dxdy

    # Perform 2D curvature analysis
    curv_data = (
        ((1 + (Cx * (dx**2))) * Cy * d2y)
        - (2 * Cx * Cy * dx * dy * dxdy)
        + ((1 + (Cy * (dy**2))) * Cx * d2x)
    ) / ((1 + (Cx * (dx**2)) + (Cy * (dy**2))) ** 1.5)

    # Rewrite attributes
    curv_data.attrs = attributes

    # Update analysis history
    hist = "2D curvature analysis performed with coefficients: {dimx}: {Cx}, {dimy}: {Cy}".format(
        dimx=dimx, Cx=Cx, dimy=dimy, Cy=Cy
    )
    curv_data.history.add(hist)

    return curv_data


def min_gradient(data, **smoothing_kwargs):
    """Perform minimum gradient analysis of data, using Gaussian filtering (see Rev. Sci. Instrum 88 (2017) 07390 for
    analysis procedure).

    Parameters
    ------------
    data : xarray.DataArray
        The data to perform minimum gradient analysis on.

    **smoothing_kwargs : float
        Axes to smooth over in the format axis=FWHM, where FWHM is the relevant FWHM of the Gaussian for convolution
        in this direction, e.g. eV=0.1. Defaults to a single pixel. Note: The use of these broadening terms is
        inequivalent to the use in the smooth function,
        i.e. data.smooth(eV=0.01).min_grad(eV=0) != data.min_grad(eV=0.01).

    Returns
    ------------
    grad_data : xarray.DataArray
        The data following minimum gradient analysis.

    Examples
    ------------
    Example usage is as follows::

        from peaks import *

        disp = load('disp.ibw')

        FM = load('FM.zip')

        # Perform minimum gradient analysis on the dispersion using Gaussian filters with FWHMs for the theta_par and eV
        # axes of 0.2 deg and 0.05 eV
        disp_grad = disp.min_gradient(theta_par=0.2, eV=0.05)

        # Perform minimum gradient analysis on the Fermi map using Gaussian filters with FWHMs for the ana_polar,
        # theta_par and eV axes of 0.2 deg, 0.2 deg and 0.05 eV
        FM_grad = FM.min_gradient(ana_polar=0.2, theta_par=0.5, eV=0.05)

    """

    # Check data is not 1D
    if len(data.dims) < 2:
        raise Exception("Function cannot act on 1D data.")

    # Copy the input data to prevent overwriting issues
    grad_data = data.copy(deep=True)

    # Define the start of the analysis history update string
    hist = "Minimum gradient analysis performed using Gaussian filters with the following FWHM along given axes: "

    # Make the sigma array (used to store standard deviations) as zeros, then update using supplied definitions
    sigma = np.zeros(len(data.dims))

    # Iterate through coordinates and determine the standard deviations in pixels from the DataArray axis scaling
    for count, value in enumerate(data.dims):
        if (
            value not in smoothing_kwargs
        ):  # No broadening given for this dimension, so set to default of a single pixel
            sigma[count] = 1 / 2.35482005
            hist += str(value) + ": 1 pixel, "  # Update analysis history string
            # Display warning that the default value has been assumed
            analysis_warning(
                "No broadening parameter supplied for {dim} dimension. Set to default of a single pixel.".format(
                    dim=value
                ),
                title="Analysis info",
                warn_type="danger",
            )
        else:  # Determine broadening in pixels from axis scaling
            delta = abs(
                data[value].data[1] - data[value].data[0]
            )  # Pixel size in relevant units for axis
            # Must convert smoothing factor from FWHM to standard deviation (in pixels)
            sigma_px = (
                np.round(smoothing_kwargs[value] / delta) / 2.35482005
            )  # Coordinate sigma in pixels
            sigma[count] = sigma_px  # Update standard deviations array
            hist += (
                str(value) + ": " + str(smoothing_kwargs[value]) + ", "
            )  # Update analysis history string
            smoothing_kwargs.pop(
                value
            )  # Remove this axis from smoothing_kwargs for consistency check later

    # Extract the raw DataArray data
    array = grad_data.data

    # Apply gradient magnitude to raw DataArray data
    array_sm = gaussian_gradient_magnitude(array, sigma)

    # Extract the renormalised gradient modulus map
    grad_data /= array_sm

    # Check that all supplied smoothing_kwargs are used, giving a warning if not (this occurs if an axis does not exist)
    if len(smoothing_kwargs) != 0:
        analysis_warning(
            "Not all supplied axes are coordinates of DataArray: {coords} have been ignored.".format(
                coords=str(smoothing_kwargs)
            ),
            title="Analysis info",
            warn_type="danger",
        )

    # Update the analysis history
    grad_data.history.add(hist[:-2])

    return grad_data



# CircularListWidget for PyQt5
# Edgar Abarca Morales
from PyQt6 import QtCore, QtWidgets


# Upgrade QtWidgets.QListWidget to support infinite arrow scrolling
class _CircularListWidget(QtWidgets.QListWidget):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def keyPressEvent(self, event):
        # Circular scrolling behaviour
        if event.key() == QtCore.Qt.Key.Key_Down:
            if self.currentRow() == self.count() - 1:
                self.setCurrentRow(0)
                return
        elif event.key() == QtCore.Qt.Key.Key_Up:
            if self.currentRow() == 0:
                self.setCurrentRow(self.count() - 1)
                return

        # Parent behavior
        super().keyPressEvent(event)



"""
Helper functions and xarray accessors for providing the user interface to metadata
"""

import copy
import pprint

import numpy as np
import pint
import pint_xarray
import xarray as xr
from pydantic import BaseModel

from peaks.core.fileIO.base_data_classes.base_data_class import BaseDataLoader
from peaks.core.utils.datatree_utils import _map_over_dt_containing_single_das

ureg = pint_xarray.unit_registry


def display_metadata(da_or_model, mode="ANSI"):
    # Recursive function to display dictionary with colored keys
    # Paul Tol's colour scheme
    colours = [
        "\x1b[38;2;187;85;0m",  # orange
        "\x1b[38;2;0;90;181m",  # blue
        "\x1b[38;2;212;17;89m",  # magenta
        "\x1b[38;2;0;133;119m",  # green
    ]
    RESET = "\x1b[0m"

    # Recursive function to display dictionary with cycling colors for each indent level
    def display_colored_dict(d, indent_level=0, col_cycle=0):
        indent = "    " * indent_level
        current_color = colours[col_cycle % len(colours)]  # Cycle through colors
        lines = []
        for key, value in d.items():
            if isinstance(value, dict):  # Nested dictionary (recursive case)
                lines.append(f"{indent}{current_color}{key}{RESET}:")
                lines.extend(
                    display_colored_dict(value, indent_level + 1, col_cycle + 1)
                )
            else:  # Base case (simple value)
                lines.append(f"{indent}{current_color}{key}{RESET}: {value}")
        return lines

    def display_colored_dict_html(d, indent_level=0, col_cycle=0):
        indent = "&nbsp;" * 4 * indent_level
        colours = ["#DDCC77", "#88CCEE", "#CC6677", "#44AA99"]
        current_color = colours[col_cycle % len(colours)]  # Cycle through colors
        lines = []
        for key, value in d.items():
            if isinstance(value, dict):  # Nested dictionary (recursive case)
                lines.append(
                    f"{indent}<span style='color:{current_color}'>{key}:</span>"
                )
                lines.extend(
                    display_colored_dict_html(value, indent_level + 1, col_cycle + 1)
                )
            else:  # Base case (simple value)
                lines.append(
                    f"{indent}<span style='color:{current_color}'>{key}:</span> {value}"
                )
        return lines

    # Display the model with colored keys
    try:
        metadata = {
            key.lstrip("_"): value.dict()
            for key, value in da_or_model.attrs.items()
            if key.startswith("_") and key not in ["_analysis_history"]
        }
    except AttributeError:
        metadata = da_or_model if isinstance(da_or_model, dict) else da_or_model.dict()
    if mode.upper() == "ANSI":
        return "\n".join(display_colored_dict(metadata))
    elif mode.upper() == "HTML":
        return "<br>".join(display_colored_dict_html(metadata))


def compare_metadata(da_or_model1, da_or_model2):
    # Function to extract metadata dictionary
    def get_metadata(da_or_model):
        try:
            metadata = {
                key.lstrip("_"): value.dict()
                for key, value in da_or_model.attrs.items()
                if key.startswith("_") and key not in ["_analysis_history"]
            }
        except AttributeError:
            metadata = da_or_model.dict()
        return metadata

    metadata1 = get_metadata(da_or_model1)
    metadata2 = get_metadata(da_or_model2)

    # Helper function to unwrap 'value' keys in dictionaries
    def unwrap_value(val):
        while isinstance(val, dict) and "value" in val and len(val) == 1:
            val = val["value"]
        return val

    # Helper function to compare values considering their types
    def values_equal(val1, val2):
        val1 = unwrap_value(val1)
        val2 = unwrap_value(val2)
        # Handle NoneType
        if val1 is None and val2 is None:
            return True
        if val1 is None or val2 is None:
            return False

        # Handle Pint quantities
        if isinstance(val1, pint.Quantity) and isinstance(val2, pint.Quantity):
            try:
                # Compare magnitudes after converting to common units
                return np.array_equal(
                    val1.to_base_units().magnitude,
                    val2.to_base_units().magnitude,
                )
            except pint.DimensionalityError:
                return False

        # Handle NumPy arrays
        if isinstance(val1, np.ndarray) and isinstance(val2, np.ndarray):
            return np.array_equal(val1, val2)

        # Handle lists or tuples
        if isinstance(val1, (list, tuple)) and isinstance(val2, (list, tuple)):
            if len(val1) != len(val2):
                return False
            return all(values_equal(v1, v2) for v1, v2 in zip(val1, val2, strict=True))

        # Handle dictionaries
        if isinstance(val1, dict) and isinstance(val2, dict):
            return compare_dicts(val1, val2) == {}

        # Default comparison
        return val1 == val2

    # Recursive function to compare two dictionaries
    def compare_dicts(d1, d2):
        differences = {}
        keys = set(d1.keys()).union(d2.keys())
        for key in keys:
            in_d1 = key in d1
            in_d2 = key in d2
            if in_d1 and in_d2:
                val1 = unwrap_value(d1[key])
                val2 = unwrap_value(d2[key])
                if isinstance(val1, dict) and isinstance(val2, dict):
                    sub_diff = compare_dicts(val1, val2)
                    if sub_diff:  # Only include if there is a difference
                        differences[key] = sub_diff
                else:
                    if not values_equal(val1, val2):
                        differences[key] = {"value1": val1, "value2": val2}
            elif in_d1:
                differences[key] = {"value1": d1[key], "value2": None}
            else:
                differences[key] = {"value1": None, "value2": d2[key]}
        return differences

    return compare_dicts(metadata1, metadata2)


@xr.register_dataset_accessor("metadata")
@xr.register_dataarray_accessor("metadata")
class Metadata:
    """Accessor for metadata on xarray DataArrays."""

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def __repr__(self):
        return display_metadata(self._obj)

    def __getattr__(self, name):
        # Return a MetadataItem for the attribute
        if f"_{name}" in self._obj.attrs:
            return MetadataItem(self._obj.attrs[f"_{name}"], path=name, obj=self._obj)
        elif name == "history":  # Add a shortcut to the history attribute
            return self._obj.history
        else:
            raise AttributeError(f"'Metadata' object has no attribute '{name}'")

    def __dir__(self):
        # Include dynamic attributes in the list of available attributes
        return super().__dir__() + list(self.keys()) + ["history"]

    def keys(self):
        return {
            k.lstrip("_"): v
            for k, v in self._obj.attrs.items()
            if k.startswith("_") and k != "_analysis_history"
        }.keys()

    # -------------- Special set methods --------------
    def set_normal_emission(self, norm_values=None, **kwargs):
        """Set the normal emission angles for the scan from a dictionary of key: value pairs to specify
        the requires angles. Like `get_normal_emission_for`, but sets the values rather than returning them.

        Parameters
        ----------
        norm_values : dict, optional
            A dictionary of the normal emission angles to set.

        **kwargs :
            Additional keyword arguments. Alternative way to pass the normal emission angles to set.
            Takes precedence over values supplied in `norm_values`.

        See Also
        --------
        get_normal_emission_for : Get the normal emission angles for the scan corresponding to a dictionary of
        key: value pairs.
        set_normal_emission_like : Set the normal emission angles for the scan to match another scan.
        assign_normal_emission : Assign the normal emission angles to a copy of the data.


        Examples
        --------
        Example usage is as follows::
            import peaks as pks

            # Load data
            FS = pks.load("path/to/data")

            # Set the normal emission angles in FS
            FS.metadata.set_normal_emission(theta_par=12, polar=5)

            # Alternatively, to return a copy of the data with the normal emission angles applied
            FS_with_norm = FS.metadata.assign_normal_emission(theta_par=12, polar=5)


        """

        # Get loc and loader class
        loc = self._obj.metadata.scan.loc
        loader = BaseDataLoader.get_loader(loc)

        if not hasattr(loader, "_parse_manipulator_references"):
            raise NotImplementedError(
                "The loader for this data does not support setting normal emission angles."
            )

        # Get and apply the new reference data to the current dataarray
        normal_emission = self.get_normal_emission_from_values(norm_values, **kwargs)
        normal_emission_dict = {
            k: {"reference_value": v} for k, v in normal_emission.items()
        }  # For passing to the metadata set methods

        self._obj.metadata.manipulator(normal_emission_dict)

    def get_normal_emission_from_values(self, norm_values=None, **kwargs):
        """Get the normal emission angles for the scan corresponding to a dictionary of key: value pairs to
        specify the required angles.

        Parameters
        ----------
        norm_values : dict
            A dictionary of the normal emission angles to set.

        **kwargs :
            Additional keyword arguments. Alternative way to pass the normal emission angles to set.
            Takes precedence over values supplied in `norm_values`.

        See Also
        --------
        set_normal_emission_like : Set the normal emission angles for the scan to match another scan.
        assign_normal_emission : Assign the normal emission angles to a copy of the data.


        Examples
        --------
        Example usage is as follows::
            import peaks as pks

            # Load data
            FS = pks.load("path/to/data")

            # Set the normal emission angles in FS
            FS.metadata.set_normal_emission(theta_par=12, polar=5)

            # Alternatively, to return a copy of the data with the normal emission angles applied
            FS_with_norm = FS.metadata.assign_normal_emission(theta_par=12, polar=5)


        """
        if norm_values is None:
            norm_values = {}
        # Update the normal emission angles with any additional kwargs
        norm_values.update(kwargs)

        # Get relevant loader class
        loc = self._obj.metadata.scan.loc
        loader = BaseDataLoader.get_loader(loc)
        if not hasattr(loader, "_parse_manipulator_references"):
            raise NotImplementedError(
                "The loader for this data does not support setting normal emission angles."
            )

        # If a pint Quantity is passed in tuple format, convert it
        for key, value in norm_values.items():
            if isinstance(value, str):
                norm_values[key] = ureg.Quantity(value)

        # Get the normal emission angles
        normal_emission = loader._parse_manipulator_references(self._obj, norm_values)

        return normal_emission

    def assign_normal_emission(self, norm_values, **kwargs):
        raise NotImplementedError("This method is not yet implemented.")

    def set_normal_emission_like(self, da):
        """Set the normal emission angles for the scan to match another scan.

        Parameters
        ----------
        da : xarray.DataArray
            The data array to match the normal emission angles to.
        """

        # Get any set reference data in da
        current_reference_data = self._get_normal_emission_dict(da)
        # Apply the new reference data to the current dataarray
        self._apply_normal_emission(self._obj, current_reference_data)

    @staticmethod
    def _apply_normal_emission(da, normal_emission_dict):
        """Apply the normal emission angles to the scan. Used as part of the `set_normal_emission_like` method.

        Parameters
        ----------
        da : xarray.DataArray
            The data array to apply the normal emission angles to.

        normal_emission_dict : dict
            A dictionary of the normal emission angles to apply.
        """
        scan_name = normal_emission_dict["scan_name"]
        normal_emission_data = {
            k: v for k, v in normal_emission_dict.items() if k != "scan_name"
        }
        if normal_emission_dict:
            # Apply the new reference data to the current dataarray
            da.metadata.manipulator(normal_emission_data)
            # Patch the analysis history
            current_history = da._analysis_history.records[-1].record
            new_history = current_history.replace(
                "were manually updated from a dictionary",
                f"were set to match the reference values from scan {scan_name}",
            )
            da._analysis_history.records[-1].record = new_history

    @staticmethod
    def _get_normal_emission_dict(da):
        """Get the normal emission angles for the scan.

        Parameters
        ----------
        da : xarray.DataArray
            The data array to get the normal emission angles of.

        Returns
        -------
        dict
            A dictionary of the normal emission angles and also scan name.
        """
        # Get the axis names in da
        axes = list(da._manipulator.dict().keys())
        # Get any set reference data in da
        current_reference_data = {
            axis: {
                "reference_value": getattr(da.metadata.manipulator, axis).reference_value
            }
            for axis in axes
            if getattr(da.metadata.manipulator, axis).reference_value is not None
        }
        current_reference_data["scan_name"] = da.name
        return current_reference_data

    def set_EF_correction(self, EF_correction):
        """Sets the Fermi level correction for a :class:`xarray.DataArray`.

        Parameters
        ----------
        EF_correction : float, int, dict or xarray.Dataset
            Fermi level correction to apply. This should be:

            - a dictionary of the form {'c0': 16.82, 'c1': -0.001, ...} specifying the coefficients of a
            polynomial fit to the Fermi edge.
            - a float or int, this will be taken as a constant shift in energy.
            - an xarray.Dataset containing the fit_result as returned by the `peaks` `.fit_gold` method

          Returns
         -------
         None
             Adds the Fermi level correction to the data attributes.
        """

        # Do some checks on the EF_correction format
        if isinstance(EF_correction, xr.Dataset):
            EF_correction = copy.deepcopy(EF_correction.attrs.get("EF_correction"))
            correction_type = "gold fit result"
        else:
            correction_type = type(EF_correction)
        if not isinstance(EF_correction, (float, int, dict)):
            raise ValueError(
                "EF_correction must be a float, int, dict of fit coefficients or xarray.Dataset containing the fit_result "
                "of the `.fit_gold` function."
            )
        if isinstance(EF_correction, dict):
            expected_keys = [f"c{i}" for i in range(len(EF_correction))]
            if not all(key in EF_correction for key in expected_keys):
                raise ValueError(
                    f"EF_correction dictionary must contain keys {expected_keys} for the polynomial fit."
                )
            for value in EF_correction.values():
                if not isinstance(value, (float, int)):
                    raise ValueError(
                        "EF_correction dictionary must contain only floats or ints as values in the form "
                        "{'c0': 16.82, 'c1': -0.001, ...} specifying the coefficients of a polynomial fit to the Fermi edge."
                    )

        self._obj.metadata.calibration.set(
            "EF_correction", copy.deepcopy(EF_correction), add_history=False
        )
        self._obj.history.add(
            f"EF_correction set to {EF_correction} from a passed {correction_type}."
        )

    def set_EF_correction_like(self, da_to_set_like):
        """Sets the Fermi level correction for a :class:`xarray.DataArray` to be the same as another DataArray.

        Parameters
        ----------
        da_to_set_like : xarray.DataArray
            DataArray to copy the Fermi level correction from.

        Returns
        -------
        None
            Adds the Fermi level correction to the data attributes.
        """
        EF_correction = copy.deepcopy(da_to_set_like.metadata.get_EF_correction())
        self._obj.metadata.calibration.set(
            "EF_correction",
            EF_correction,
            add_history=False,
        )
        self._obj.history.add(
            f"EF_correction set to {EF_correction} to match scan `{da_to_set_like.name}`."
        )

    def get_EF_correction(self):
        """Get the Fermi level correction for a :class:`xarray.DataArray`.

        Returns
        -------
        dict or float or int
            Fermi level correction dictionary.
        """
        EF_correction = self._obj.metadata.calibration.EF_correction
        if isinstance(EF_correction, MetadataItem):
            return EF_correction.__dict__["_data"]
        return EF_correction


@xr.register_datatree_accessor("metadata")
class MetadataDT:
    """Accessor for metadata on xarray DataTrees."""

    def __init__(self, xarray_obj):
        self._obj = xarray_obj

    def __call__(self, metadata_dict=None, **kwargs):
        if metadata_dict is None:
            metadata_dict = {}
        metadata_dict.update(kwargs)

        def apply_metadata_to_ds(ds):
            # Apply the metadata to the dataset if it exists at dataset level
            if len(ds.metadata.keys()) > 0:
                for key, value in metadata_dict.items():
                    getattr(ds.metadata, key)(value)
            else:
                # Otherwise map over all dataarrays in the dataset
                for key, value in metadata_dict.items():
                    ds.map(
                        lambda da, key=key, value=value: (
                            getattr(da.metadata, key)(value),
                            da,
                        )[1],
                        keep_attrs=False,
                    )

            return ds

        for _key, _value in metadata_dict.items():
            self._obj.map_over_datasets(apply_metadata_to_ds)

        return self

    def set_normal_emission(self):
        raise NotImplementedError("This method is not yet implemented.")

    def set_normal_emission_like(self, da):
        # Get any set reference data in da
        current_reference_data = Metadata._get_normal_emission_dict(da)

        def apply_normal_to_ds(ds):
            # Apply the metadata to the dataset if it exists at dataset level
            if len(ds.metadata.keys()) > 0:
                Metadata._apply_normal_emission(ds, current_reference_data)
            else:
                # Otherwise map over all dataarrays in the dataset
                ds.map(
                    lambda da: (
                        Metadata._apply_normal_emission(da, current_reference_data) or da
                    )
                )
            return ds

        self._obj.map_over_datasets(apply_normal_to_ds)

    def set_EF_correction(self, EF_correction):
        _map_over_dt_containing_single_das(
            self._obj, lambda da: da.metadata.set_EF_correction(EF_correction)
        )

    def set_EF_correction_like(self, da_to_set_like):
        _map_over_dt_containing_single_das(
            self._obj, lambda da: da.metadata.set_EF_correction_like(da_to_set_like)
        )

    set_EF_correction.__doc__ = Metadata.set_EF_correction.__doc__
    set_EF_correction_like.__doc__ = Metadata.set_EF_correction_like.__doc__


class MetadataItem:
    """Class to handle metadata items."""

    def __init__(self, data, path="", obj=None):
        self._data = data
        self._path = path  # Path to the current attribute
        self._obj = obj  # Reference to the parent object

    def __repr__(self):
        return display_metadata(self._data)

    def __getattr__(self, name):
        if name in (
            "_data",
            "_path",
            "_obj",
            "__class__",
            "__dict__",
            "__weakref__",
            "__module__",
        ):
            return object.__getattribute__(self, name)
        data = self._data
        new_path = f"{self._path}.{name}" if self._path else name
        if isinstance(data, BaseModel):
            if hasattr(data, name):
                attr = getattr(data, name)
                if isinstance(attr, (BaseModel, dict)):
                    return MetadataItem(
                        attr, path=new_path, obj=self._obj
                    )  # Pass self._obj
                else:
                    return attr  # Return the value directly
            else:
                raise AttributeError(
                    f"'{data.__class__.__name__}' object has no attribute '{name}'"
                )
        elif isinstance(data, dict):
            if name in data:
                attr = data[name]
                if isinstance(attr, (BaseModel, dict)):
                    return MetadataItem(
                        attr, path=new_path, obj=self._obj
                    )  # Pass self._obj
                else:
                    return attr  # Return the value directly
            else:
                raise AttributeError(f"Dict has no key '{name}'")
        else:
            raise AttributeError(
                f"Cannot access attribute '{name}' on type {type(data)}"
            )

    def __setattr__(self, name, value):
        if name in ("_data", "_path", "_obj"):
            object.__setattr__(self, name, value)
        else:
            self.set(name, value)

    def set(self, name, value, add_history=True):
        data = self._data
        full_path = f"{self._path}.{name}" if self._path else name
        if isinstance(data, BaseModel):
            if hasattr(data, name):
                current_value = getattr(data, name)
                if isinstance(current_value, pint.Quantity) and not isinstance(
                    value, pint.Quantity
                ):
                    new_value = value * current_value.units
                else:
                    new_value = value
                setattr(data, name, new_value)
                if add_history:
                    self._obj.history.add(
                        f"Metadata attribute '{full_path}' was manually set to {new_value}",
                        ".metadata",
                    )
            else:
                raise AttributeError(
                    f"'{data.__class__.__name__}' object has no attribute '{name}'"
                )
        elif isinstance(data, dict):
            current_value = data.get(name)
            if isinstance(current_value, pint.Quantity) and not isinstance(
                value, pint.Quantity
            ):
                new_value = value * current_value.units
            else:
                new_value = value
            data[name] = new_value
            if add_history:
                self._obj.history.add(
                    f"Metadata attribute '{full_path}' was manually set to {new_value}",
                    ".metadata",
                )
        else:
            raise AttributeError(f"Cannot set attribute '{name}' on type {type(data)}")

    def __call__(self, value=None):
        if value is not None:
            data = self._data
            path = self._path
            if isinstance(data, BaseModel):
                if isinstance(value, dict):
                    self._update_model(data, value, path)
                else:
                    raise ValueError(
                        "Value must be a dictionary to update the pydantic model."
                    )
            elif isinstance(data, dict):
                if isinstance(value, dict):
                    self._update_dict(data, value, path)
                else:
                    raise ValueError("Value must be a dictionary to update the dict.")
            else:
                raise TypeError(
                    "Underlying data is neither a pydantic model nor a dict."
                )
            # After all updates, add a single history entry
            if self._obj is not None and hasattr(self._obj, "history"):
                # Format the passed dictionary for readability
                value_str = pprint.pformat(value)
                self._obj.history.add(
                    f"Metadata attributes for '{path}' were manually updated from a dictionary: {value_str}",
                    ".metadata",
                )
        else:
            print("No value provided to update the metadata.")
        return self

    def _update_model(self, model, updates, path):
        for key, value in updates.items():
            if hasattr(model, key):
                current_value = getattr(model, key)
                full_path = f"{path}.{key}" if path else key
                if isinstance(current_value, BaseModel):
                    if isinstance(value, dict):
                        self._update_model(current_value, value, full_path)
                    else:
                        raise ValueError(
                            f"Expected dict for updating '{key}', got {type(value)}"
                        )
                else:
                    if isinstance(current_value, pint.Quantity) and not isinstance(
                        value, pint.Quantity
                    ):
                        new_value = value * current_value.units
                    else:
                        new_value = value
                    setattr(model, key, new_value)
            else:
                raise AttributeError(
                    f"'{model.__class__.__name__}' object has no attribute '{key}'"
                )

    def _update_dict(self, data_dict, updates, path):
        for key, value in updates.items():
            current_value = data_dict.get(key)
            full_path = f"{path}.{key}" if path else key
            if isinstance(current_value, dict):
                if isinstance(value, dict):
                    self._update_dict(current_value, value, full_path)
                else:
                    raise ValueError(
                        f"Expected dict for updating '{key}', got {type(value)}"
                    )
            else:
                if isinstance(current_value, pint.Quantity) and not isinstance(
                    value, pint.Quantity
                ):
                    new_value = value * current_value.units
                else:
                    new_value = value
                data_dict[key] = new_value

    def __dir__(self):
        data = self._data
        if isinstance(data, BaseModel):
            return super().__dir__() + list(data.__fields__.keys())
        elif isinstance(data, dict):
            return super().__dir__() + list(data.keys())
        else:
            return super().__dir__()



import pint_xarray

from peaks.core.fileIO.base_arpes_data_classes.base_specs_class import SpecsDataLoader
from peaks.core.fileIO.loc_registry import register_loader

ureg = pint_xarray.unit_registry


@register_loader
class StAndrewsSpecs(SpecsDataLoader):
    """Data loader for St Andrews Specs ARPES system, measured using Specs Prodigy.

    Define _scan_axis_resolution_order to define preferences for the primary dimension of a
    3D scan where more than one user axis varies."""

    _loc_name = "StA_Phoibos"
    _loc_description = "SPECS Phoibos 225 in King Group at St Andrews"
    _loc_url = "https://www.quantummatter.co.uk/king"
    _analyser_slit_angle = 90 * ureg.deg
    _scan_axis_resolution_order = ["tilt", "polar"]
    _manipulator_name_conventions = {
        "polar": "Theta",
        "tilt": "Phi",
        "azi": "Azi",
        "x1": "Y",
        "x2": "Z",
        "x3": "X",
    }
    _temperature_exclude_from_metadata_warn = [
        "sample",
        "cryostat",
        "shield",
        "setpoint",
    ]
    _manipulator_exclude_from_metadata_warn = [
        axis for axis in _manipulator_name_conventions.keys()
    ]

    _manipulator_sign_conventions = {
        "azi": -1,
        "x1": -1,
        "x2": -1,
        "x3": -1,
    }
    _analyser_sign_conventions = {}

    _SPECS_metadata_key_mappings = {}
    _SPECS_metadata_units = {}

    @classmethod
    def _load_data(cls, fpath, lazy):
        data_dict = super()._load_data(fpath, lazy)

        if "azi" in data_dict["coords"] and "tilt" in data_dict["coords"]:
            # The azi should be a secondary axis to the tilt, and should be removed
            new_coords = {
                dim: coord for dim, coord in data_dict["coords"].items() if dim != "azi"
            }
            data_dict["coords"] = new_coords

        return data_dict
